"""Unit tests for pyatv.interface."""

import unittest

from pyatv import (convert, exceptions, interface)
from pyatv.const import MediaType, DeviceState, RepeatState, ShuffleState


class TestClass:

    variable = 1234

    def test_method(self):
        """Help text."""
        pass

    def another_method(self):
        """Some other help text. More text here. Test."""
        pass

    @property
    def some_property(self):
        """Property help"""
        pass

    def abbrev_help(self):
        """Type, e.g. a, b or c."""
        pass

    def abbrev_help_more_text(self):
        """Type, e.g. a, b or c. Some other text."""
        pass

    def _private_method_ignored(self):
        """Not parsed."""
        pass


class PlayingDummy(interface.Playing):

    def __init__(self, media_type=MediaType.Video,
                 device_state=DeviceState.Playing, title=None,
                 artist=None, album=None, genre=None, total_time=None,
                 position=None, shuffle=ShuffleState.Songs,
                 repeat=RepeatState.Track):
        self._media_type = media_type
        self._device_state = device_state
        self._title = title
        self._artist = artist
        self._album = album
        self._genre = genre
        self._total_time = total_time
        self._position = position
        self._shuffle = shuffle
        self._repeat = repeat

    @property
    def media_type(self):
        """What type of media is currently playing, e.g. video, music."""
        return self._media_type

    @property
    def device_state(self):
        """Current device state, e.g. playing or paused."""
        return self._device_state

    @property
    def title(self):
        """Title of the current media, e.g. movie or song name."""
        return self._title

    @property
    def artist(self):
        """Artist of the currently playing song."""
        return self._artist

    @property
    def album(self):
        """Album of the currently playing song."""
        return self._album

    @property
    def genre(self):
        """Genre of the currently playing song."""
        return self._genre

    @property
    def total_time(self):
        """Total play time in seconds."""
        return self._total_time

    @property
    def position(self):
        """Current position in the playing media (seconds)."""
        return self._position

    @property
    def shuffle(self):
        """If shuffle is enabled or not."""
        return self._shuffle

    @property
    def repeat(self):
        """Current repeat mode."""
        return self._repeat


class MetadataDummy(interface.Metadata):

    def artwork(self):
        """Return artwork for what is currently playing (or None)."""
        raise exceptions.NotSupportedError()

    def artwork_id(self):
        """Return a unique identifier for current artwork."""
        raise exceptions.NotSupportedError()

    def playing(self):
        """Return what is currently playing."""
        raise exceptions.NotSupportedError()


class InterfaceTest(unittest.TestCase):

    def setUp(self):
        self.methods = interface.retrieve_commands(TestClass)

    # COMMANDS AND HELP TEXT

    def test_get_commands(self):
        self.assertEqual(5, len(self.methods))
        self.assertTrue('test_method' in self.methods)
        self.assertTrue('another_method' in self.methods)
        self.assertTrue('some_property' in self.methods)
        self.assertTrue('abbrev_help' in self.methods)

    def test_get_first_sentence_without_leading_period_in_pydoc(self):
        self.assertEqual('Help text', self.methods['test_method'])
        self.assertEqual(
            'Some other help text', self.methods['another_method'])
        self.assertEqual('Property help', self.methods['some_property'])

    def test_try_to_be_smart_with_abbreviations(self):
        self.assertEqual(
            'Type, e.g. a, b or c', self.methods['abbrev_help'])
        self.assertEqual(
            'Type, e.g. a, b or c', self.methods['abbrev_help_more_text'])

    # PLAYING STR

    def test_playing_media_type_and_playstate(self):
        out = str(PlayingDummy())
        self.assertIn(convert.media_type_str(MediaType.Video), out)
        self.assertIn(convert.playstate_str(DeviceState.Playing), out)

    def test_playing_title_artist_album_genre(self):
        out = str(PlayingDummy(
            title='mytitle', artist='myartist',
            album='myalbum', genre='mygenre'))
        self.assertIn('mytitle', out)
        self.assertIn('myartist', out)
        self.assertIn('myalbum', out)
        self.assertIn('mygenre', out)

    def test_playing_only_position(self):
        self.assertIn('1234', str(PlayingDummy(position=1234)))

    def test_playing_only_total_time(self):
        self.assertIn('5678', str(PlayingDummy(total_time=5678)))

    def test_playing_both_position_and_total_time(self):
        out = str(PlayingDummy(position=1234, total_time=5678))
        self.assertIn('1234/5678', out)

    def test_playing_shuffle_and_repeat(self):
        out = str(PlayingDummy())
        self.assertIn('Shuffle: Songs', out)
        self.assertIn('Repeat: Track', out)

    # OTHER

    def test_playing_generate_same_hash(self):
        playing = PlayingDummy(
            title='title', artist='artist', album='album', total_time=123)
        self.assertEqual(
            '538df531d1715629fdd87affd0c5957bcbf54cd89180778071e6535b7df4e22c',
            playing.hash)

        playing2 = PlayingDummy(
            title='dummy', artist='test', album='none', total_time=321)
        self.assertEqual(
            '80045c05d18382f33a5369fd5cdfc6ae42c3eb418125f638d7a31ab173b01ade',
            playing2.hash)

    # METADATA

    def test_metadata_device_id(self):
        self.assertEqual(MetadataDummy('dummy').device_id, 'dummy')

    # Dummy test for the sake of code coverage
    def test_metadata_rest_not_supported(self):
        metadata = MetadataDummy('dummy')

        with self.assertRaises(exceptions.NotSupportedError):
            metadata.artwork()
        with self.assertRaises(exceptions.NotSupportedError):
            metadata.artwork_id()
        with self.assertRaises(exceptions.NotSupportedError):
            metadata.playing()



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
)

from google.protobuf.internal.containers import (
    RepeatedScalarFieldContainer as google___protobuf___internal___containers___RepeatedScalarFieldContainer,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    Iterable as typing___Iterable,
    Optional as typing___Optional,
    Text as typing___Text,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class LanguageOption(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    type = ... # type: int
    languageTag = ... # type: typing___Text
    characteristics = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]
    displayName = ... # type: typing___Text
    identifier = ... # type: typing___Text

    def __init__(self,
        *,
        type : typing___Optional[int] = None,
        languageTag : typing___Optional[typing___Text] = None,
        characteristics : typing___Optional[typing___Iterable[typing___Text]] = None,
        displayName : typing___Optional[typing___Text] = None,
        identifier : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> LanguageOption: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"displayName",u"identifier",u"languageTag",u"type"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"characteristics",u"displayName",u"identifier",u"languageTag",u"type"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"displayName",b"displayName",u"identifier",b"identifier",u"languageTag",b"languageTag",u"type",b"type"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"characteristics",b"characteristics",u"displayName",b"displayName",u"identifier",b"identifier",u"languageTag",b"languageTag",u"type",b"type"]) -> None: ...



"""Constants used in the public API."""

from enum import Enum


MAJOR_VERSION = '0'
MINOR_VERSION = '4'
PATCH_VERSION = '0a13'
__short_version__ = '{}.{}'.format(MAJOR_VERSION, MINOR_VERSION)
__version__ = '{}.{}'.format(__short_version__, PATCH_VERSION)


class Protocol(Enum):
    """All supported protocol."""

    DMAP = 1
    MRP = 2
    AirPlay = 3


class MediaType(Enum):
    """All suuported media types."""

    Unknown = 0
    Video = 1
    Music = 2
    TV = 3


class DeviceState(Enum):
    """All supported device states."""

    Idle = 0
    Loading = 1
    Paused = 2
    Playing = 3
    Stopped = 4
    Seeking = 5


class RepeatState(Enum):
    """All supported repeat states."""

    Off = 0
    Track = 1
    All = 2


class ShuffleState(Enum):
    """All supported shuffle states."""

    Off = 0
    Albums = 1
    Songs = 2



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    FieldDescriptor as google___protobuf___descriptor___FieldDescriptor,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class CryptoPairingMessage(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    pairingData = ... # type: bytes
    status = ... # type: int
    isRetrying = ... # type: bool
    isUsingSystemPairing = ... # type: bool
    state = ... # type: int

    def __init__(self,
        *,
        pairingData : bytes,
        status : int,
        isRetrying : bool,
        isUsingSystemPairing : bool,
        state : int,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> CryptoPairingMessage: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"isRetrying",u"isUsingSystemPairing",u"pairingData",u"state",u"status"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"isRetrying",u"isUsingSystemPairing",u"pairingData",u"state",u"status"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"isRetrying",b"isRetrying",u"isUsingSystemPairing",b"isUsingSystemPairing",u"pairingData",b"pairingData",u"state",b"state",u"status",b"status"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"isRetrying",b"isRetrying",u"isUsingSystemPairing",b"isUsingSystemPairing",u"pairingData",b"pairingData",u"state",b"state",u"status",b"status"]) -> None: ...

cryptoPairingMessage = ... # type: google___protobuf___descriptor___FieldDescriptor



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    FieldDescriptor as google___protobuf___descriptor___FieldDescriptor,
)

from google.protobuf.internal.containers import (
    RepeatedScalarFieldContainer as google___protobuf___internal___containers___RepeatedScalarFieldContainer,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    Iterable as typing___Iterable,
    Optional as typing___Optional,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class SendCommandResultMessage(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    errorCode = ... # type: int
    handlerReturnStatus = ... # type: int
    handlerReturnStatusDatas = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[bytes]

    def __init__(self,
        *,
        errorCode : typing___Optional[int] = None,
        handlerReturnStatus : typing___Optional[int] = None,
        handlerReturnStatusDatas : typing___Optional[typing___Iterable[bytes]] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> SendCommandResultMessage: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"errorCode",u"handlerReturnStatus"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"errorCode",u"handlerReturnStatus",u"handlerReturnStatusDatas"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"errorCode",b"errorCode",u"handlerReturnStatus",b"handlerReturnStatus"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"errorCode",b"errorCode",u"handlerReturnStatus",b"handlerReturnStatus",u"handlerReturnStatusDatas",b"handlerReturnStatusDatas"]) -> None: ...

sendCommandResultMessage = ... # type: google___protobuf___descriptor___FieldDescriptor



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pyatv/mrp/protobuf/LanguageOption.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pyatv/mrp/protobuf/LanguageOption.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\'pyatv/mrp/protobuf/LanguageOption.proto\"u\n\x0eLanguageOption\x12\x0c\n\x04type\x18\x01 \x01(\x05\x12\x13\n\x0blanguageTag\x18\x02 \x01(\t\x12\x17\n\x0f\x63haracteristics\x18\x03 \x03(\t\x12\x13\n\x0b\x64isplayName\x18\x04 \x01(\t\x12\x12\n\nidentifier\x18\x05 \x01(\t')
)




_LANGUAGEOPTION = _descriptor.Descriptor(
  name='LanguageOption',
  full_name='LanguageOption',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='LanguageOption.type', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='languageTag', full_name='LanguageOption.languageTag', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='characteristics', full_name='LanguageOption.characteristics', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='displayName', full_name='LanguageOption.displayName', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='identifier', full_name='LanguageOption.identifier', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=43,
  serialized_end=160,
)

DESCRIPTOR.message_types_by_name['LanguageOption'] = _LANGUAGEOPTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LanguageOption = _reflection.GeneratedProtocolMessageType('LanguageOption', (_message.Message,), {
  'DESCRIPTOR' : _LANGUAGEOPTION,
  '__module__' : 'pyatv.mrp.protobuf.LanguageOption_pb2'
  # @@protoc_insertion_point(class_scope:LanguageOption)
  })
_sym_db.RegisterMessage(LanguageOption)


# @@protoc_insertion_point(module_scope)



"""Functional tests using the API with a fake Apple TV."""

import pyatv
import ipaddress
import asynctest

from unittest.mock import patch

from pyatv.const import Protocol
from tests import fake_udns, zeroconf_stub


IP_1 = '10.0.0.1'
IP_2 = '10.0.0.2'
IP_3 = '10.0.0.3'
IP_4 = '10.0.0.4'
IP_5 = '10.0.0.5'
IP_6 = '10.0.0.6'
IP_LOCALHOST = '127.0.0.1'

HSGID = 'hsgid'

MRP_ID_1 = 'mrp_id_1'
MRP_ID_2 = 'mrp_id_2'

AIRPLAY_ID = 'AA:BB:CC:DD:EE:FF'

HOMESHARING_SERVICE_1 = zeroconf_stub.homesharing_service(
    'AAAA', b'Apple TV 1', IP_1, b'aaaa')
HOMESHARING_SERVICE_2 = zeroconf_stub.homesharing_service(
    'BBBB', b'Apple TV 2', IP_2, b'bbbb')
HOMESHARING_SERVICE_3 = zeroconf_stub.homesharing_service(
    'CCCC', b'Apple TV 3', IP_3, b'cccc')
DEVICE_SERVICE_1 = zeroconf_stub.device_service(
    'CCCC', b'Apple TV 3', IP_3)
MRP_SERVICE_1 = zeroconf_stub.mrp_service(
    'DDDD', b'Apple TV 4', IP_4, MRP_ID_1)
MRP_SERVICE_2 = zeroconf_stub.mrp_service(
    'EEEE', b'Apple TV 5', IP_5, MRP_ID_2)
AIRPLAY_SERVICE_1 = zeroconf_stub.airplay_service(
    'Apple TV 6', IP_6, AIRPLAY_ID)


def _get_atv(atvs, ip):
    for atv in atvs:
        if atv.address == ipaddress.ip_address(ip):
            return atv
    return None


class ScanZeroconfTest(asynctest.TestCase):

    async def test_scan_no_device_found(self):
        zeroconf_stub.stub(pyatv)

        atvs = await pyatv.scan(self.loop, timeout=0)
        self.assertEqual(len(atvs), 0)

    async def test_scan(self):
        zeroconf_stub.stub(
            pyatv, HOMESHARING_SERVICE_1, HOMESHARING_SERVICE_2,
            MRP_SERVICE_1, AIRPLAY_SERVICE_1)

        atvs = await pyatv.scan(self.loop, timeout=0)
        self.assertEqual(len(atvs), 4)

        # First device
        dev1 = _get_atv(atvs, IP_1)
        self.assertIsNotNone(dev1)
        self.assertEqual(dev1.identifier, 'AAAA')

        # Second device
        dev2 = _get_atv(atvs, IP_2)
        self.assertIsNotNone(dev2)
        self.assertEqual(dev2.identifier, 'BBBB')

        # Third device
        dev3 = _get_atv(atvs, IP_4)
        self.assertIsNotNone(dev3)
        self.assertEqual(dev3.identifier, MRP_ID_1)

        # Fourth device
        dev4 = _get_atv(atvs, IP_6)
        self.assertIsNotNone(dev4)
        self.assertEqual(dev4.identifier, AIRPLAY_ID)

    async def test_scan_no_home_sharing(self):
        zeroconf_stub.stub(pyatv, DEVICE_SERVICE_1)

        atvs = await pyatv.scan(self.loop, timeout=0)
        self.assertEqual(len(atvs), 1)
        self.assertEqual(atvs[0].name, 'Apple TV 3')
        self.assertEqual(atvs[0].address, ipaddress.ip_address(IP_3))

        atv = atvs[0]
        self.assertEqual(
            atv.get_service(Protocol.DMAP).port, 3689)

    async def test_scan_home_sharing_merge(self):
        zeroconf_stub.stub(pyatv, DEVICE_SERVICE_1, HOMESHARING_SERVICE_3)

        atvs = await pyatv.scan(self.loop, timeout=0)
        self.assertEqual(len(atvs), 1)
        self.assertEqual(atvs[0].name, 'Apple TV 3')
        self.assertEqual(atvs[0].address, ipaddress.ip_address('10.0.0.3'))

        service = atvs[0].main_service()
        self.assertEqual(service.credentials, 'cccc')
        self.assertEqual(service.port, 3689)

    async def test_scan_mrp(self):
        zeroconf_stub.stub(
          pyatv, MRP_SERVICE_1, MRP_SERVICE_2, DEVICE_SERVICE_1)

        atvs = await pyatv.scan(
          self.loop, timeout=0, protocol=Protocol.MRP)
        self.assertEqual(len(atvs), 2)

        dev1 = _get_atv(atvs, IP_4)
        self.assertIsNotNone(dev1)
        self.assertEqual(dev1.name, 'Apple TV 4')
        self.assertIsNotNone(dev1.get_service(Protocol.MRP))

        dev2 = _get_atv(atvs, IP_5)
        self.assertIsNotNone(dev2)
        self.assertEqual(dev2.name, 'Apple TV 5')
        self.assertIsNotNone(dev2.get_service(Protocol.MRP))

    async def test_scan_airplay_device(self):
        zeroconf_stub.stub(pyatv, AIRPLAY_SERVICE_1)

        atvs = await pyatv.scan(self.loop, timeout=0)
        self.assertEqual(len(atvs), 1)
        self.assertEqual(atvs[0].name, 'Apple TV 6')
        self.assertEqual(atvs[0].address, ipaddress.ip_address('10.0.0.6'))

        services = atvs[0].services
        self.assertEqual(len(services), 1)

        service = services[0]
        self.assertEqual(service.port, 7000)

    async def test_scan_for_particular_device(self):
        zeroconf_stub.stub(pyatv, HOMESHARING_SERVICE_1, HOMESHARING_SERVICE_2)

        atvs = await pyatv.scan(
            self.loop, timeout=0, identifier='BBBB')
        self.assertEqual(len(atvs), 1)
        self.assertEqual(atvs[0].name, 'Apple TV 2')
        self.assertEqual(atvs[0].address, ipaddress.ip_address(IP_2))


class ScanUnicastDnsTest(asynctest.TestCase):

    async def setUp(self, *services):
        self.server = fake_udns.FakeUdns(self.loop)
        await self.server.start()

    async def scan(self):
        port = str(self.server.port)
        with patch.dict('os.environ', {'PYATV_UDNS_PORT': port}):
            return await pyatv.scan(self.loop, hosts=[IP_LOCALHOST])

    def assertDevice(self, atv, name, address,
                     identifier, protocol, port, creds=None):
        self.assertEqual(atv.name, name)
        self.assertEqual(atv.address, address)
        self.assertEqual(atv.identifier, identifier)
        self.assertIsNotNone(atv.get_service(protocol))
        self.assertEqual(atv.get_service(protocol).port, port)
        self.assertEqual(atv.get_service(protocol).credentials, creds)

    async def test_scan_no_results(self):
        atvs = await self.scan()
        self.assertEqual(len(atvs), 0)

    async def test_missing_port(self):
        self.server.add_service(
            fake_udns.mrp_service('dummy', 'dummy', 'dummy', port=None))

        atvs = await self.scan()
        self.assertEqual(len(atvs), 0)

    async def test_missing_properties(self):
        self.server.add_service(
            fake_udns.FakeDnsService('dummy', 1234, None))

        atvs = await self.scan()
        self.assertEqual(len(atvs), 0)

    async def test_scan_mrp(self):
        self.server.add_service(
            fake_udns.mrp_service('Apple TV', 'Apple TV MRP', MRP_ID_1))

        atvs = await self.scan()
        self.assertEqual(len(atvs), 1)

        self.assertDevice(
            atvs[0], 'Apple TV MRP', IP_LOCALHOST,
            MRP_ID_1, Protocol.MRP, 49152)

    async def test_scan_airplay(self):
        self.server.add_service(
            fake_udns.airplay_service('Apple TV', AIRPLAY_ID))

        atvs = await self.scan()
        self.assertEqual(len(atvs), 1)

        self.assertDevice(
            atvs[0], 'Apple TV', IP_LOCALHOST,
            AIRPLAY_ID, Protocol.AirPlay, 7000)

    async def test_scan_homesharing(self):
        self.server.add_service(
            fake_udns.homesharing_service('abcd', 'Apple TV HS', HSGID))

        atvs = await self.scan()
        self.assertEqual(len(atvs), 1)

        self.assertDevice(
            atvs[0], 'Apple TV HS', IP_LOCALHOST,
            'abcd', Protocol.DMAP, 3689, HSGID)

    async def test_scan_no_homesharing(self):
        self.server.add_service(
            fake_udns.device_service('Apple TV', 'Apple TV Device'))

        atvs = await self.scan()
        self.assertEqual(len(atvs), 1)

        self.assertDevice(
            atvs[0], 'Apple TV Device', IP_LOCALHOST,
            'Apple TV', Protocol.DMAP, 3689)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pyatv/mrp/protobuf/RegisterHIDDeviceMessage.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pyatv.mrp.protobuf import ProtocolMessage_pb2 as pyatv_dot_mrp_dot_protobuf_dot_ProtocolMessage__pb2
from pyatv.mrp.protobuf import VirtualTouchDeviceDescriptorMessage_pb2 as pyatv_dot_mrp_dot_protobuf_dot_VirtualTouchDeviceDescriptorMessage__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pyatv/mrp/protobuf/RegisterHIDDeviceMessage.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n1pyatv/mrp/protobuf/RegisterHIDDeviceMessage.proto\x1a(pyatv/mrp/protobuf/ProtocolMessage.proto\x1a<pyatv/mrp/protobuf/VirtualTouchDeviceDescriptorMessage.proto\"S\n\x18RegisterHIDDeviceMessage\x12\x37\n\x10\x64\x65viceDescriptor\x18\x01 \x01(\x0b\x32\x1d.VirtualTouchDeviceDescriptor:M\n\x18registerHIDDeviceMessage\x12\x10.ProtocolMessage\x18\x0b \x01(\x0b\x32\x19.RegisterHIDDeviceMessage')
  ,
  dependencies=[pyatv_dot_mrp_dot_protobuf_dot_ProtocolMessage__pb2.DESCRIPTOR,pyatv_dot_mrp_dot_protobuf_dot_VirtualTouchDeviceDescriptorMessage__pb2.DESCRIPTOR,])


REGISTERHIDDEVICEMESSAGE_FIELD_NUMBER = 11
registerHIDDeviceMessage = _descriptor.FieldDescriptor(
  name='registerHIDDeviceMessage', full_name='registerHIDDeviceMessage', index=0,
  number=11, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR)


_REGISTERHIDDEVICEMESSAGE = _descriptor.Descriptor(
  name='RegisterHIDDeviceMessage',
  full_name='RegisterHIDDeviceMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='deviceDescriptor', full_name='RegisterHIDDeviceMessage.deviceDescriptor', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=157,
  serialized_end=240,
)

_REGISTERHIDDEVICEMESSAGE.fields_by_name['deviceDescriptor'].message_type = pyatv_dot_mrp_dot_protobuf_dot_VirtualTouchDeviceDescriptorMessage__pb2._VIRTUALTOUCHDEVICEDESCRIPTOR
DESCRIPTOR.message_types_by_name['RegisterHIDDeviceMessage'] = _REGISTERHIDDEVICEMESSAGE
DESCRIPTOR.extensions_by_name['registerHIDDeviceMessage'] = registerHIDDeviceMessage
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RegisterHIDDeviceMessage = _reflection.GeneratedProtocolMessageType('RegisterHIDDeviceMessage', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERHIDDEVICEMESSAGE,
  '__module__' : 'pyatv.mrp.protobuf.RegisterHIDDeviceMessage_pb2'
  # @@protoc_insertion_point(class_scope:RegisterHIDDeviceMessage)
  })
_sym_db.RegisterMessage(RegisterHIDDeviceMessage)

registerHIDDeviceMessage.message_type = _REGISTERHIDDEVICEMESSAGE
pyatv_dot_mrp_dot_protobuf_dot_ProtocolMessage__pb2.ProtocolMessage.RegisterExtension(registerHIDDeviceMessage)

# @@protoc_insertion_point(module_scope)



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    FieldDescriptor as google___protobuf___descriptor___FieldDescriptor,
)

from google.protobuf.internal.containers import (
    RepeatedScalarFieldContainer as google___protobuf___internal___containers___RepeatedScalarFieldContainer,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    Iterable as typing___Iterable,
    Optional as typing___Optional,
    Text as typing___Text,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class NotificationMessage(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    notification = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]
    userInfo = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[bytes]

    def __init__(self,
        *,
        notification : typing___Optional[typing___Iterable[typing___Text]] = None,
        userInfo : typing___Optional[typing___Iterable[bytes]] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> NotificationMessage: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"notification",u"userInfo"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"notification",b"notification",u"userInfo",b"userInfo"]) -> None: ...

notificationMessage = ... # type: google___protobuf___descriptor___FieldDescriptor



"""Stub for the zeroconf library.

As zeroconf does not provide a stub or mock, this implementation will serve as
stub here. It can fake immediate answers for any service.
"""
from aiozeroconf import ServiceInfo


def homesharing_service(service_name, atv_name, address, hsgid):
    """Create a new home sharing service simulating an Apple TV."""
    # Mostly default values here for now
    props = {
        b'DFID': b'2', b'PrVs': b'65538', b'hG': hsgid, b'Name': atv_name,
        b'txtvers': b'1', b'atSV': b'65541', b'MiTPV': b'196611',
        b'EiTS': b'1', b'fs': b'2', b'MniT': b'167845888'
    }

    return ServiceInfo('_appletv-v2._tcp.local.',
                       service_name + '._appletv-v2._tcp.local.',
                       address=address, port=3689,
                       properties=props)


def mrp_service(service_name, atv_name, address, identifier, port=49152):
    """Create a MediaRemote service simulating an Apple TV."""
    props = {
        b'ModelName': b'Mac', b'SystemBuildVersion': b'16G29',
        b'Name': atv_name, b'AllowPairing': b'YES',
        b'UniqueIdentifier': identifier.encode('utf-8')
    }

    return ServiceInfo('_mediaremotetv._tcp.local.',
                       service_name + '._mediaremotetv._tcp.local.',
                       address=address, port=port,
                       properties=props)


def airplay_service(atv_name, address, deviceid, port=7000):
    """Create a MediaRemote service simulating an Apple TV."""
    props = {
        b'deviceid': deviceid.encode('utf-8'), b'model': b'AppleTV3,1',
        b'pi': b'4EE5AF58-7E5D-465A-935E-82E4DB74385D', b'flags': b'0x44',
        b'vv': b'2', b'features': b'0x5A7FFFF7,0xE',
        b'pk': b'3853c0e2ce3844727ca0cb1b86a3e3875e66924d2648d8f8caf71f8118793d98',  # noqa
        b'srcvers': b'220.68'
    }

    return ServiceInfo('_airplay._tcp.local.',
                       atv_name + '._airplay._tcp.local.',
                       address=address, port=port,
                       properties=props)


def device_service(service_name, atv_name, address):
    """Create a service representing an Apple TV with no home-sharing."""
    props = {
        b'DvTy': b'AppleTV', b'Ver': b'131077', b'DvSv': b'1792',
        b'atCV': b'65539', b'atSV': b'65541', b'txtvers': b'1',
        b'DbId': b'AAAAAAAAAAAAAAAA', b'CtlN': atv_name
    }

    return ServiceInfo('_touch-able._tcp.local.',
                       service_name + '._touch-able._tcp.local.',
                       address=address, port=3689,
                       server='AppleTV-2.local.',
                       properties=props)


class ServiceBrowserStub:
    """Stub for ServiceBrowser."""

    def __init__(self, zeroconf, service_type, listener):
        """Create a new instance of ServiceBrowser."""
        for service in zeroconf.services:
            if service.type == service_type:
                listener.add_service(zeroconf, service_type, service.name)


class ZeroconfStub:
    """Stub for Zeroconf."""

    def __init__(self, services):
        """Create a new instance of Zeroconf."""
        self.services = services
        self.registered_services = []

    async def get_service_info(self, service_type, service_name, timeout=2000):
        """Look up service information."""
        for service in self.services:
            if service.name == service_name:
                return service

    async def register_service(self, service):
        """Save services registered services."""
        self.registered_services.append(service)

    async def unregister_service(self, service):
        """Stub for unregistering services (does nothing)."""
        pass

    async def close(self):
        """Stub for closing zeroconf (does nothing)."""
        pass


def stub(module, *services):
    """Stub a module using zeroconf."""
    instance = ZeroconfStub(list(services))
    module.Zeroconf = lambda loop, address_family=None: instance
    module.ServiceBrowser = ServiceBrowserStub
    return instance



"""Functional tests using the API with a fake DMAP Apple TV."""

import asyncio
import ipaddress

from aiohttp.test_utils import unittest_run_loop

from pyatv import connect, exceptions
from pyatv.conf import (AirPlayService, DmapService, AppleTV)
from pyatv.const import ShuffleState
from pyatv.dmap import pairing
from tests.dmap.fake_dmap_atv import (FakeAppleTV, AppleTVUseCases)
from tests.airplay.fake_airplay_device import DEVICE_CREDENTIALS
from tests import (zeroconf_stub, common_functional_tests)
from tests.common_functional_tests import DummyDeviceListener

HSGID = '12345-6789-0'
PAIRING_GUID = '0x0000000000000001'
SESSION_ID = 55555
REMOTE_NAME = 'pyatv remote'
PIN_CODE = 1234

ARTWORK_BYTES = b'1234'
ARTWORK_MIMETYPE = 'image/png'
AIRPLAY_STREAM = 'http://stream'

# This is valid for the PAIR in the pairing module and pin 1234
# (extracted form a real device)
PAIRINGCODE = '690E6FF61E0D7C747654A42AED17047D'

HOMESHARING_SERVICE_1 = zeroconf_stub.homesharing_service(
    'AAAA', b'Apple TV 1', '10.0.0.1', b'aaaa')
HOMESHARING_SERVICE_2 = zeroconf_stub.homesharing_service(
    'BBBB', b'Apple TV 2', '10.0.0.2', b'bbbb')


class DummyPushListener:

    @staticmethod
    def playstatus_update(updater, playstatus):
        updater.stop()

    @staticmethod
    def playstatus_error(updater, exception):
        pass


class DMAPFunctionalTest(common_functional_tests.CommonFunctionalTests):

    async def setUpAsync(self):
        await super().setUpAsync()
        self.atv = await self.get_connected_device(HSGID)

        # TODO: currently stubs internal method, should provide stub
        # for netifaces later
        pairing._get_private_ip_addresses = \
            lambda: [ipaddress.ip_address('10.0.0.1')]

    async def tearDownAsync(self):
        await self.atv.close()
        await super().tearDownAsync()

    async def get_application(self, loop=None):
        self.fake_atv = FakeAppleTV(
            HSGID, PAIRING_GUID, SESSION_ID, self)
        self.usecase = AppleTVUseCases(self.fake_atv)
        return self.fake_atv.app

    async def get_connected_device(self, hsgid):
        self.dmap_service = DmapService(
            'dmap_id', hsgid, port=self.server.port)
        self.airplay_service = AirPlayService(
            'airplay_id', self.server.port, DEVICE_CREDENTIALS)
        self.conf = AppleTV('127.0.0.1', 'Apple TV')
        self.conf.add_service(self.dmap_service)
        self.conf.add_service(self.airplay_service)
        return await connect(self.conf, self.loop)

    @unittest_run_loop
    async def test_not_supportedt(self):
        with self.assertRaises(exceptions.NotSupportedError):
            await self.atv.remote_control.suspend()

    @unittest_run_loop
    async def test_connect_failed(self):
        # Twice since the client will retry one time
        self.usecase.make_login_fail()
        self.usecase.make_login_fail()

        with self.assertRaises(exceptions.AuthenticationError):
            await self.atv.connect()

    # This test verifies issue #2 (automatic re-login). It uses the artwork
    # API, but it could have been any API since the login code is the same.
    @unittest_run_loop
    async def test_relogin_if_session_expired(self):
        await self.atv.connect()

        # Here, we are logged in and currently have a asession id. These
        # usescases will result in being logged out (HTTP 403) and forcing a
        # re-login with a new session id (1234)
        self.usecase.example_video()
        self.usecase.force_relogin(1234)
        self.usecase.artwork_no_permission()
        self.usecase.change_artwork(ARTWORK_BYTES, ARTWORK_MIMETYPE)

        artwork = await self.atv.metadata.artwork()
        self.assertEqual(artwork.bytes, ARTWORK_BYTES)

    @unittest_run_loop
    async def test_login_with_hsgid_succeed(self):
        session_id = await self.atv.connect()
        self.assertEqual(SESSION_ID, session_id)

    @unittest_run_loop
    async def test_login_with_pairing_guid_succeed(self):
        await self.atv.close()
        self.atv = await self.get_connected_device(PAIRING_GUID)
        session_id = await self.atv.connect()
        self.assertEqual(SESSION_ID, session_id)

    @unittest_run_loop
    async def test_connection_lost(self):
        self.usecase.server_closes_connection()

        self.atv.listener = DummyDeviceListener()
        self.atv.push_updater.listener = DummyPushListener()
        self.atv.push_updater.start()

        # Callback is scheduled on the event loop, so a semaphore is used
        # to synchronize with the loop
        await asyncio.wait_for(
            self.atv.listener.lost_sem.acquire(), timeout=3.0)

    @unittest_run_loop
    async def test_button_unsupported_raises(self):
        buttons = ['home', 'volume_up', 'volume_down', 'suspend', 'wakeup']
        for button in buttons:
            with self.assertRaises(exceptions.NotSupportedError):
                await getattr(self.atv.remote_control, button)()

    @unittest_run_loop
    async def test_shuffle_state_albums(self):
        # DMAP does not support "albums" as shuffle state, so it is
        # mapped to "songs"
        self.usecase.example_video(shuffle=ShuffleState.Albums)
        playing = await self.playing(shuffle=ShuffleState.Songs)
        self.assertEqual(playing.shuffle, ShuffleState.Songs)

    @unittest_run_loop
    async def test_set_shuffle_albums(self):
        self.usecase.example_video()

        # DMAP does not support "albums" as shuffle state, so it is
        # mapped to "songs"
        await self.atv.remote_control.set_shuffle(ShuffleState.Albums)
        playing = await self.playing(shuffle=ShuffleState.Songs)
        self.assertEqual(playing.shuffle, ShuffleState.Songs)

    @unittest_run_loop
    async def test_play_url_no_service(self):
        conf = AppleTV('127.0.0.1', 'Apple TV')
        conf.add_service(self.dmap_service)

        atv = await connect(conf, self.loop)

        with self.assertRaises(exceptions.NotSupportedError):
            await atv.stream.play_url('http://123')

        await atv.close()



"""Play media on a device by sending an URL."""

import logging
import asyncio
import plistlib
from uuid import uuid4

from pyatv import exceptions


_LOGGER = logging.getLogger(__name__)

PLAY_RETRIES = 3
WAIT_RETRIES = 5
TIMEOUT = 10
HEADERS = {
    'User-Agent': 'MediaControl/1.0',
    'Content-Type': 'application/x-apple-binary-plist'
}


# pylint: disable=too-few-public-methods
class AirPlayPlayer:
    """This class helps with playing media from an URL."""

    def __init__(self, loop, http):
        """Initialize a new AirPlay instance."""
        self.loop = loop
        self.http = http

    async def play_url(self, url, position=0):
        """Play media from an URL on the device."""
        body = {
            'Content-Location': url,
            'Start-Position': position,
            'X-Apple-Session-ID': str(uuid4()),
            }

        retry = 0
        while retry < PLAY_RETRIES:
            # pylint: disable=no-member
            _, status = await self.http.post_data(
                'play',
                headers=HEADERS,
                data=plistlib.dumps(body, fmt=plistlib.FMT_BINARY),
                timeout=TIMEOUT)

            # Sometimes AirPlay fails with "Internal Server Error", we
            # apply a "lets try again"-approach to that
            if status == 500:
                retry += 1
                _LOGGER.debug('Failed to stream %s, retry %d of %d',
                              url, retry, PLAY_RETRIES)
                await asyncio.sleep(1.0, loop=self.loop)
                continue

            # TODO: Should be more fine-grained
            if 400 <= status < 600:
                raise exceptions.AuthenticationError(
                    'Status code: ' + str(status))

            await self._wait_for_media_to_end()
            return

        raise exceptions.PlaybackError('Max retries exceeded')

    # Poll playback-info to find out if something is playing. It might take
    # some time until the media starts playing, give it 5 seconds (attempts)
    async def _wait_for_media_to_end(self):
        attempts = WAIT_RETRIES
        video_started = False

        while True:
            data, status = await self.http.get_data('playback-info')

            if status == 403:
                raise exceptions.NoCredentialsError(
                    'device authentication required')

            _LOGGER.debug('Playback-info (%d): %s', status, data)

            parsed = plistlib.loads(data)

            # duration is only available if something is playing
            if 'duration' in parsed:
                video_started = True
                attempts = -1
            else:
                video_started = False
                if attempts >= 0:
                    attempts -= 1

            if not video_started and attempts < 0:
                _LOGGER.debug('media playback ended')
                break

            await asyncio.sleep(1, loop=self.loop)



"""Simple example that connects to a device with autodiscover."""

from pyatv import helpers


# Method that is dispatched by the asyncio event loop
async def print_what_is_playing(atv):
    """Print what is playing for the discovered device."""
    playing = await atv.metadata.playing()
    print('Currently playing:')
    print(playing)


# logout is automatically performed by auto_connect
helpers.auto_connect(print_what_is_playing)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pyatv/mrp/protobuf/TransactionPackets.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pyatv.mrp.protobuf import TransactionPacket_pb2 as pyatv_dot_mrp_dot_protobuf_dot_TransactionPacket__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pyatv/mrp/protobuf/TransactionPackets.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n+pyatv/mrp/protobuf/TransactionPackets.proto\x1a*pyatv/mrp/protobuf/TransactionPacket.proto\"9\n\x12TransactionPackets\x12#\n\x07packets\x18\x01 \x03(\x0b\x32\x12.TransactionPacket')
  ,
  dependencies=[pyatv_dot_mrp_dot_protobuf_dot_TransactionPacket__pb2.DESCRIPTOR,])




_TRANSACTIONPACKETS = _descriptor.Descriptor(
  name='TransactionPackets',
  full_name='TransactionPackets',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='packets', full_name='TransactionPackets.packets', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=91,
  serialized_end=148,
)

_TRANSACTIONPACKETS.fields_by_name['packets'].message_type = pyatv_dot_mrp_dot_protobuf_dot_TransactionPacket__pb2._TRANSACTIONPACKET
DESCRIPTOR.message_types_by_name['TransactionPackets'] = _TRANSACTIONPACKETS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TransactionPackets = _reflection.GeneratedProtocolMessageType('TransactionPackets', (_message.Message,), {
  'DESCRIPTOR' : _TRANSACTIONPACKETS,
  '__module__' : 'pyatv.mrp.protobuf.TransactionPackets_pb2'
  # @@protoc_insertion_point(class_scope:TransactionPackets)
  })
_sym_db.RegisterMessage(TransactionPackets)


# @@protoc_insertion_point(module_scope)



"""Fake DMAP Apple TV device for tests."""

import logging

from collections import namedtuple

from aiohttp import web

from pyatv.dmap import (parser, tags, tag_definitions)
from tests.airplay.fake_airplay_device import (
    FakeAirPlayDevice, AirPlayUseCases)
from tests import utils

_LOGGER = logging.getLogger(__name__)

EXPECTED_HEADERS = {
    'Accept': '*/*',
    'Accept-Encoding': 'gzip',
    'Client-DAAP-Version': '3.13',
    'Client-ATV-Sharing-Version': '1.2',
    'Client-iTunes-Sharing-Version': '3.15',
    'User-Agent': 'Remote/1021',
    'Viewer-Only-Client': '1',
}

LoginResponse = namedtuple('LoginResponse', 'session, status')
ArtworkResponse = namedtuple('ArtworkResponse', 'content, status')
AirPlayPlaybackResponse = namedtuple('AirPlayPlaybackResponse', 'content')
PairingResponse = namedtuple('PairingResponse', 'remote_name, pairing_code')


DEVICE_IDENTIFIER = '75FBEEC773CFC563'
DEVICE_AUTH_KEY = \
    '8F06696F2542D70DF59286C761695C485F815BE3D152849E1361282D46AB1493'
DEVICE_PIN = 2271
DEVICE_CREDENTIALS = DEVICE_IDENTIFIER + ':' + DEVICE_AUTH_KEY


class PlayingResponse:
    """Response returned by command playstatusupdate."""

    def __init__(self, revision=0, shuffle=False, **kwargs):
        """Initialize a new PlayingResponse."""
        self.paused = kwargs.get('paused', None)
        self.title = kwargs.get('title', None)
        self.artist = kwargs.get('artist', None)
        self.album = kwargs.get('album', None)
        self.genre = kwargs.get('genre', None)
        self.total_time = kwargs.get('total_time', None)
        self.position = kwargs.get('position', None)
        self.mediakind = kwargs.get('mediakind', None)
        self.playstatus = kwargs.get('playstatus', None)
        self.repeat = kwargs.get('repeat', None)
        self.revision = revision
        self.shuffle = shuffle
        self.force_close = kwargs.get('force_close', False)


class FakeAppleTV(FakeAirPlayDevice):
    """Implementation of a fake DMAP Apple TV."""

    def __init__(self, hsgid, pairing_guid, session_id, testcase):
        """Initialize a new FakeAppleTV."""
        super().__init__(testcase)

        self.responses['login'] = [LoginResponse(session_id, 200)]
        self.responses['artwork'] = []
        self.responses['playing'] = []
        self.responses['pairing'] = []
        self.hsgid = hsgid
        self.pairing_guid = pairing_guid
        self.session = None
        self.last_button_pressed = None
        self.buttons_press_count = 0

        self.app.router.add_get('/login', self.handle_login)
        self.app.router.add_get(
            '/ctrl-int/1/playstatusupdate', self.handle_playstatus)
        self.app.router.add_post(
            '/ctrl-int/1/controlpromptentry', self.handle_remote_button)
        self.app.router.add_get(
            '/ctrl-int/1/nowplayingartwork', self.handle_artwork)
        self.app.router.add_post(
            '/ctrl-int/1/setproperty', self.handle_set_property)
        for button in ['play', 'pause', 'stop', 'nextitem', 'previtem']:
            self.app.router.add_post('/ctrl-int/1/' + button,
                                     self.handle_playback_button)

    async def handle_login(self, request):
        """Handle login requests."""
        self._verify_headers(request)
        self._verify_auth_parameters(
            request, check_login_id=True, check_session=False)

        data = self._get_response('login')
        mlid = tags.uint32_tag('mlid', data.session)
        mlog = tags.container_tag('mlog', mlid)
        self.session = data.session
        return web.Response(body=mlog, status=data.status)

    async def handle_playback_button(self, request):
        """Handle playback buttons."""
        self._verify_auth_parameters(request)
        self.last_button_pressed = request.rel_url.path.split('/')[-1]
        self.buttons_press_count += 1
        return web.Response(status=200)

    async def handle_remote_button(self, request):
        """Handle remote control buttons."""
        self._verify_auth_parameters(request)
        content = await request.content.read()
        parsed = parser.parse(content, tag_definitions.lookup_tag)
        self.last_button_pressed = self._convert_button(parsed)
        self.buttons_press_count += 1
        return web.Response(status=200)

    def _convert_button(self, data):
        value = parser.first(data, 'cmbe')

        # Consider navigation buttons if six commands have been received
        if self.buttons_press_count == 6:
            if value == 'touchUp&time=6&point=20,250':
                return 'up'
            elif value == 'touchUp&time=6&point=20,275':
                return 'down'
            elif value == 'touchUp&time=7&point=50,100':
                return 'left'
            elif value == 'touchUp&time=7&point=75,100':
                return 'right'

        return value

    async def handle_artwork(self, request):
        """Handle artwork requests."""
        self._verify_auth_parameters(request)
        artwork = self._get_response('artwork')
        return web.Response(body=artwork.content, status=artwork.status)

    async def handle_playstatus(self, request):
        """Handle  playstatus (currently playing) requests."""
        self._verify_auth_parameters(request)

        body = b''
        playing = self._get_response('playing')

        # Check if connection should be closed to trigger error on client side
        if playing.force_close:
            await request.transport.close()

        # Make sure revision matches
        revision = int(request.rel_url.query['revision-number'])
        if playing.revision != revision:
            # Not a valid response as a real device, just to make tests fail
            return web.Response(status=500)

        if playing.paused is not None:
            body += tags.uint32_tag('caps', 3 if playing.paused else 4)

        if playing.title is not None:
            body += tags.string_tag('cann', playing.title)

        if playing.artist is not None:
            body += tags.string_tag('cana', playing.artist)

        if playing.album is not None:
            body += tags.string_tag('canl', playing.album)

        if playing.genre is not None:
            body += tags.string_tag('cang', playing.genre)

        if playing.total_time is not None:
            total_time = playing.total_time * 1000  # sec -> ms
            body += tags.uint32_tag('cast', total_time)

            if playing.position is not None:
                pos = (playing.total_time - playing.position)
                print(playing.total_time, playing.position)
                body += tags.uint32_tag('cant', pos * 1000)  # sec -> ms

        if playing.mediakind is not None:
            body += tags.uint32_tag('cmmk', playing.mediakind)

        if playing.playstatus is not None:
            body += tags.uint32_tag('caps', playing.playstatus)

        if playing.repeat is not None:
            body += tags.uint8_tag('carp', playing.repeat)

        body += tags.uint8_tag('cash', playing.shuffle)
        body += tags.uint32_tag('cmsr', playing.revision + 1)

        return web.Response(
            body=tags.container_tag('cmst', body), status=200)

    async def handle_set_property(self, request):
        """Handle property changes."""
        self._verify_auth_parameters(request)
        if 'dacp.playingtime' in request.rel_url.query:
            playtime = int(request.rel_url.query['dacp.playingtime'])
            self._get_response('playing').position = int(playtime / 1000)
        elif 'dacp.shufflestate' in request.rel_url.query:
            shuffle = int(request.rel_url.query['dacp.shufflestate'])
            self._get_response('playing').shuffle = (shuffle == 1)
        elif 'dacp.repeatstate' in request.rel_url.query:
            repeat = int(request.rel_url.query['dacp.repeatstate'])
            self._get_response('playing').repeat = repeat
        else:
            web.Response(body=b'', status=500)

        return web.Response(body=b'', status=200)

    async def trigger_bonjour(self, stubbed_zeroconf):
        """Act upon available Bonjour services."""
        # Add more services here when needed
        for service in stubbed_zeroconf.registered_services:
            if service.type != '_touch-remote._tcp.local.':
                continue

            # Look for the response matching this remote
            remote_name = service.properties[b'DvNm']
            for response in self.responses['pairing']:
                if response.remote_name == remote_name:
                    return await self.perform_pairing(response, service.port)

    async def perform_pairing(self, pairing_response, port):
        """Pair with a remote client.

        This will perform a GET-request to the specified port and hand over
        information to the client (pyatv) so that the pairing process can be
        completed.
        """
        server = 'http://127.0.0.1:{}'.format(port)
        url = '{}/pairing?pairingcode={}&servicename=test'.format(
            server, pairing_response.pairing_code)
        data, _ = await utils.simple_get(url)

        # Verify content returned in pairingresponse
        parsed = parser.parse(data, tag_definitions.lookup_tag)
        self.tc.assertEqual(parser.first(parsed, 'cmpa', 'cmpg'), 1)
        self.tc.assertEqual(parser.first(parsed, 'cmpa', 'cmnm'),
                            pairing_response.remote_name)
        self.tc.assertEqual(parser.first(parsed, 'cmpa', 'cmty'), 'iPhone')

    # Verifies that all needed headers are included in the request. Should be
    # checked in all requests, but that seems a bit too much and not that
    # necessary.
    def _verify_headers(self, request):
        for header in EXPECTED_HEADERS:
            self.tc.assertIn(header, request.headers)
            self.tc.assertEqual(request.headers[header],
                                EXPECTED_HEADERS[header])

    # This method makes sure that the correct login id and/or session id is
    # included in the GET-parameters. As this is extremely important for
    # anything to work, this should be verified in all requests.
    def _verify_auth_parameters(self,
                                request,
                                check_login_id=False,
                                check_session=True):
        params = request.rel_url.query

        # Either hsgid or pairing-guid should be present
        if check_login_id:
            if 'hsgid' in params:
                self.tc.assertEqual(params['hsgid'], self.hsgid,
                                    msg='hsgid does not match')
            elif 'pairing-guid' in params:
                self.tc.assertEqual(params['pairing-guid'], self.pairing_guid,
                                    msg='pairing-guid does not match')
            else:
                self.tc.assertTrue(False, 'hsgid or pairing-guid not found')

        if check_session:
            self.tc.assertEqual(int(params['session-id']), self.session,
                                msg='session id does not match')


class AppleTVUseCases(AirPlayUseCases):
    """Wrapper for altering behavior of a FakeAppleTV instance.

    Extend and use this class to alter behavior of a fake Apple TV device.
    """

    def __init__(self, fake_apple_tv):
        """Initialize a new AppleTVUseCases."""
        self.device = fake_apple_tv

    def force_relogin(self, session):
        """Call this method to change current session id."""
        self.device.responses['login'].append(LoginResponse(session, 200))

    def make_login_fail(self, immediately=True):
        """Call this method to make login fail with response 503."""
        response = LoginResponse(0, 503)
        if immediately:
            self.device.responses['login'].append(response)
        else:
            self.device.responses['login'].insert(0, response)

    def change_artwork(self, artwork, mimetype, identifier=None):
        """Call this method to change artwork response."""
        self.device.responses['artwork'].append(
            ArtworkResponse(artwork, 200))

    def artwork_no_permission(self):
        """Make artwork fail with no permission.

        This corresponds to have been logged out for some reason.
        """
        self.device.responses['artwork'].insert(
            0, ArtworkResponse(None, 403))

    def nothing_playing(self):
        """Call this method to put device in idle state."""
        self.device.responses['playing'].insert(0, PlayingResponse())

    def server_closes_connection(self):
        """Call this method to force server to close connection on request."""
        self.device.responses['playing'].insert(
            0, PlayingResponse(force_close=True))

    def example_video(self, **kwargs):
        """Play some example video."""
        self.video_playing(paused=True, title='dummy',
                           total_time=123, position=3, **kwargs)

    def video_playing(self, paused, title, total_time, position, **kwargs):
        """Call this method to change what is currently plaing to video."""
        revision = 0
        shuffle = False
        repeat = None
        if 'revision' in kwargs:
            revision = kwargs['revision']
        if 'shuffle' in kwargs:
            shuffle = kwargs['shuffle'].value
        if 'repeat' in kwargs:
            repeat = kwargs['repeat'].value
        self.device.responses['playing'].insert(0, PlayingResponse(
            revision=revision,
            paused=paused, title=title,
            total_time=total_time, position=position,
            mediakind=3, shuffle=shuffle, repeat=repeat))

    def music_playing(self, paused, artist, album, title, genre,
                      total_time, position):
        """Call this method to change what is currently plaing to music."""
        self.device.responses['playing'].insert(0, PlayingResponse(
            paused=paused, title=title,
            artist=artist, album=album,
            total_time=total_time,
            position=position, genre=genre,
            mediakind=2))

    def media_is_loading(self):
        """Call this method to put device in a loading state."""
        self.device.responses['playing'].insert(0, PlayingResponse(
            playstatus=1))

    def pairing_response(self, remote_name, expected_pairing_code):
        """Reponse when a pairing request is made."""
        self.device.responses['pairing'].insert(0, PairingResponse(
            remote_name, expected_pairing_code))

    async def act_on_bonjour_services(self, stubbed_zeroconf):
        """Act on available Bonjour services.

        This will make the device look at published services and perform
        actions base on these. Most imporant for the pairing process.
        """
        await self.device.trigger_bonjour(stubbed_zeroconf)



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    EnumDescriptor as google___protobuf___descriptor___EnumDescriptor,
)

from google.protobuf.internal.containers import (
    RepeatedScalarFieldContainer as google___protobuf___internal___containers___RepeatedScalarFieldContainer,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    Iterable as typing___Iterable,
    List as typing___List,
    Optional as typing___Optional,
    Tuple as typing___Tuple,
    cast as typing___cast,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class TextInputTraits(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class AutocapitalizationType(int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: int) -> str: ...
        @classmethod
        def Value(cls, name: str) -> TextInputTraits.AutocapitalizationType: ...
        @classmethod
        def keys(cls) -> typing___List[str]: ...
        @classmethod
        def values(cls) -> typing___List[TextInputTraits.AutocapitalizationType]: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[str, TextInputTraits.AutocapitalizationType]]: ...
        NONE = typing___cast(TextInputTraits.AutocapitalizationType, 0)
        WORDS = typing___cast(TextInputTraits.AutocapitalizationType, 1)
        SENTENCES = typing___cast(TextInputTraits.AutocapitalizationType, 2)
        CHARACTERS = typing___cast(TextInputTraits.AutocapitalizationType, 3)
    NONE = typing___cast(TextInputTraits.AutocapitalizationType, 0)
    WORDS = typing___cast(TextInputTraits.AutocapitalizationType, 1)
    SENTENCES = typing___cast(TextInputTraits.AutocapitalizationType, 2)
    CHARACTERS = typing___cast(TextInputTraits.AutocapitalizationType, 3)

    class KeyboardType(int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: int) -> str: ...
        @classmethod
        def Value(cls, name: str) -> TextInputTraits.KeyboardType: ...
        @classmethod
        def keys(cls) -> typing___List[str]: ...
        @classmethod
        def values(cls) -> typing___List[TextInputTraits.KeyboardType]: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[str, TextInputTraits.KeyboardType]]: ...
        KEYBOARD_TYPE_DEFAULT = typing___cast(TextInputTraits.KeyboardType, 0)
        ASCII_CAPABLE = typing___cast(TextInputTraits.KeyboardType, 1)
        NUMBERS_AND_PUNCTUATION = typing___cast(TextInputTraits.KeyboardType, 2)
        URL = typing___cast(TextInputTraits.KeyboardType, 3)
        NUMBER_PAD = typing___cast(TextInputTraits.KeyboardType, 4)
        PHONE_PAD = typing___cast(TextInputTraits.KeyboardType, 5)
        NAME_PHONE_PAD = typing___cast(TextInputTraits.KeyboardType, 6)
        EMAIL_ADDRESS = typing___cast(TextInputTraits.KeyboardType, 7)
        DECIMAL_PAD = typing___cast(TextInputTraits.KeyboardType, 8)
        TWITTER = typing___cast(TextInputTraits.KeyboardType, 9)
        WEB_SEARCH = typing___cast(TextInputTraits.KeyboardType, 10)
    KEYBOARD_TYPE_DEFAULT = typing___cast(TextInputTraits.KeyboardType, 0)
    ASCII_CAPABLE = typing___cast(TextInputTraits.KeyboardType, 1)
    NUMBERS_AND_PUNCTUATION = typing___cast(TextInputTraits.KeyboardType, 2)
    URL = typing___cast(TextInputTraits.KeyboardType, 3)
    NUMBER_PAD = typing___cast(TextInputTraits.KeyboardType, 4)
    PHONE_PAD = typing___cast(TextInputTraits.KeyboardType, 5)
    NAME_PHONE_PAD = typing___cast(TextInputTraits.KeyboardType, 6)
    EMAIL_ADDRESS = typing___cast(TextInputTraits.KeyboardType, 7)
    DECIMAL_PAD = typing___cast(TextInputTraits.KeyboardType, 8)
    TWITTER = typing___cast(TextInputTraits.KeyboardType, 9)
    WEB_SEARCH = typing___cast(TextInputTraits.KeyboardType, 10)

    class ReturnKeyType(int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: int) -> str: ...
        @classmethod
        def Value(cls, name: str) -> TextInputTraits.ReturnKeyType: ...
        @classmethod
        def keys(cls) -> typing___List[str]: ...
        @classmethod
        def values(cls) -> typing___List[TextInputTraits.ReturnKeyType]: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[str, TextInputTraits.ReturnKeyType]]: ...
        RETURN_KEY_DEFAULT = typing___cast(TextInputTraits.ReturnKeyType, 0)
        GO = typing___cast(TextInputTraits.ReturnKeyType, 1)
        GOOGLE = typing___cast(TextInputTraits.ReturnKeyType, 2)
        JOIN = typing___cast(TextInputTraits.ReturnKeyType, 3)
        NEXT = typing___cast(TextInputTraits.ReturnKeyType, 4)
        ROUTE = typing___cast(TextInputTraits.ReturnKeyType, 5)
        SEARCH = typing___cast(TextInputTraits.ReturnKeyType, 6)
        SEND = typing___cast(TextInputTraits.ReturnKeyType, 7)
        YAHOO = typing___cast(TextInputTraits.ReturnKeyType, 8)
        DONE = typing___cast(TextInputTraits.ReturnKeyType, 9)
        EMERGENCY_CALL = typing___cast(TextInputTraits.ReturnKeyType, 10)
        CONTINUE = typing___cast(TextInputTraits.ReturnKeyType, 11)
    RETURN_KEY_DEFAULT = typing___cast(TextInputTraits.ReturnKeyType, 0)
    GO = typing___cast(TextInputTraits.ReturnKeyType, 1)
    GOOGLE = typing___cast(TextInputTraits.ReturnKeyType, 2)
    JOIN = typing___cast(TextInputTraits.ReturnKeyType, 3)
    NEXT = typing___cast(TextInputTraits.ReturnKeyType, 4)
    ROUTE = typing___cast(TextInputTraits.ReturnKeyType, 5)
    SEARCH = typing___cast(TextInputTraits.ReturnKeyType, 6)
    SEND = typing___cast(TextInputTraits.ReturnKeyType, 7)
    YAHOO = typing___cast(TextInputTraits.ReturnKeyType, 8)
    DONE = typing___cast(TextInputTraits.ReturnKeyType, 9)
    EMERGENCY_CALL = typing___cast(TextInputTraits.ReturnKeyType, 10)
    CONTINUE = typing___cast(TextInputTraits.ReturnKeyType, 11)

    autocapitalizationType = ... # type: TextInputTraits.AutocapitalizationType
    keyboardType = ... # type: TextInputTraits.KeyboardType
    returnKeyType = ... # type: TextInputTraits.ReturnKeyType
    autocorrection = ... # type: bool
    spellchecking = ... # type: bool
    enablesReturnKeyAutomatically = ... # type: bool
    secureTextEntry = ... # type: bool
    validTextRangeLocation = ... # type: int
    validTextRangeLength = ... # type: int
    PINEntrySeparatorIndexes = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[int]

    def __init__(self,
        *,
        autocapitalizationType : typing___Optional[TextInputTraits.AutocapitalizationType] = None,
        keyboardType : typing___Optional[TextInputTraits.KeyboardType] = None,
        returnKeyType : typing___Optional[TextInputTraits.ReturnKeyType] = None,
        autocorrection : typing___Optional[bool] = None,
        spellchecking : typing___Optional[bool] = None,
        enablesReturnKeyAutomatically : typing___Optional[bool] = None,
        secureTextEntry : typing___Optional[bool] = None,
        validTextRangeLocation : typing___Optional[int] = None,
        validTextRangeLength : typing___Optional[int] = None,
        PINEntrySeparatorIndexes : typing___Optional[typing___Iterable[int]] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> TextInputTraits: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"autocapitalizationType",u"autocorrection",u"enablesReturnKeyAutomatically",u"keyboardType",u"returnKeyType",u"secureTextEntry",u"spellchecking",u"validTextRangeLength",u"validTextRangeLocation"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"PINEntrySeparatorIndexes",u"autocapitalizationType",u"autocorrection",u"enablesReturnKeyAutomatically",u"keyboardType",u"returnKeyType",u"secureTextEntry",u"spellchecking",u"validTextRangeLength",u"validTextRangeLocation"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"autocapitalizationType",b"autocapitalizationType",u"autocorrection",b"autocorrection",u"enablesReturnKeyAutomatically",b"enablesReturnKeyAutomatically",u"keyboardType",b"keyboardType",u"returnKeyType",b"returnKeyType",u"secureTextEntry",b"secureTextEntry",u"spellchecking",b"spellchecking",u"validTextRangeLength",b"validTextRangeLength",u"validTextRangeLocation",b"validTextRangeLocation"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"PINEntrySeparatorIndexes",b"PINEntrySeparatorIndexes",u"autocapitalizationType",b"autocapitalizationType",u"autocorrection",b"autocorrection",u"enablesReturnKeyAutomatically",b"enablesReturnKeyAutomatically",u"keyboardType",b"keyboardType",u"returnKeyType",b"returnKeyType",u"secureTextEntry",b"secureTextEntry",u"spellchecking",b"spellchecking",u"validTextRangeLength",b"validTextRangeLength",u"validTextRangeLocation",b"validTextRangeLocation"]) -> None: ...



"""Definitions of DMAP tags used by various applications."""

import logging
from .parser import DmapTag
from .tags import (
    read_bool, read_uint, read_str, read_ignore, read_bplist)

_LOGGER = logging.getLogger(__name__)


# Internal version that works like read_ignore, but also logs
def _read_unknown(data, start, length):
    _LOGGER.warning('Unknown data: %s', str(data[start-8:start+length+8]))


# These are the tags that we know about so far
# pylint: disable=bad-whitespace
_TAGS = {
    'aeFP': DmapTag(read_uint,   'com.apple.itunes.req-fplay'),
    'aeSV': DmapTag(read_uint,   'com.apple.itunes.music-sharing-version'),
    'apro': DmapTag(read_uint,   'daap.protocolversion'),
    'asgr': DmapTag(read_uint,   'com.apple.itunes.gapless-resy'),
    'ated': DmapTag(read_bool,   'daap.supportsextradata'),
    'caar': DmapTag(read_uint,   'dacp.albumrepeat'),
    'caas': DmapTag(read_uint,   'dacp.albumshuffle'),
    'caci': DmapTag('container', 'dacp.controlint'),
    'cafe': DmapTag(read_bool,   'dacp.fullscreenenabled'),
    'cafs': DmapTag(read_uint,   'dacp.fullscreen'),
    'cana': DmapTag(read_str,    'daap.nowplayingartist'),
    'cang': DmapTag(read_str,    'dacp.nowplayinggenre'),
    'canl': DmapTag(read_str,    'daap.nowplayingalbum'),
    'cann': DmapTag(read_str,    'daap.nowplayingtrack'),
    'cant': DmapTag(read_uint,   'dacp.remainingtime'),
    'capr': DmapTag(read_uint,   'dacp.protocolversion'),
    'caps': DmapTag(read_uint,   'dacp.playstatus'),
    'carp': DmapTag(read_uint,   'dacp.repeatstate'),
    'cash': DmapTag(read_uint,   'dacp.shufflestate'),
    'cast': DmapTag(read_uint,   'dacp.tracklength'),
    'casu': DmapTag(read_uint,   'dacp.su'),
    'cavc': DmapTag(read_bool,   'dacp.volumecontrollable'),
    'cave': DmapTag(read_bool,   'dacp.dacpvisualizerenabled'),
    'cavs': DmapTag(read_uint,   'dacp.visualizer'),
    'ceQR': DmapTag('container', 'com.apple.itunes.playqueue-contents-response'),  # NOQA
    'ceSD': DmapTag(read_bplist, 'playing metadata'),
    'cmcp': DmapTag('container', 'dmcp.controlprompt'),
    'cmmk': DmapTag(read_uint,   'dmcp.mediakind'),
    'cmnm': DmapTag(read_str,    'dacp.devicename'),
    'cmpa': DmapTag('container', 'dacp.pairinganswer'),
    'cmpg': DmapTag(read_uint,   'dacp.pairingguid'),
    'cmpr': DmapTag(read_uint,   'dmcp.protocolversion'),
    'cmsr': DmapTag(read_uint,   'dmcp.serverrevision'),
    'cmst': DmapTag('container', 'dmcp.playstatus'),
    'cmty': DmapTag(read_str,    'dacp.devicetype'),
    'mdcl': DmapTag('container', 'dmap.dictionary'),
    'miid': DmapTag(read_uint,   'dmap.itemid'),
    'minm': DmapTag(read_str,    'dmap.itemname'),
    'mlcl': DmapTag('container', 'dmap.listing'),
    'mlid': DmapTag(read_uint,   'dmap.sessionid'),
    'mlit': DmapTag('container', 'dmap.listingitem'),
    'mlog': DmapTag('container', 'dmap.loginresponse'),
    'mpro': DmapTag(read_uint,   'dmap.protocolversion'),
    'mrco': DmapTag(read_uint,   'dmap.returnedcount'),
    'msal': DmapTag(read_bool,   'dmap.supportsautologout'),
    'msbr': DmapTag(read_bool,   'dmap.supportsbrowse'),
    'msdc': DmapTag(read_uint,   'dmap.databasescount'),
    'msed': DmapTag(read_bool,   'dmap.supportsedit'),
    'msex': DmapTag(read_bool,   'dmap.supportsextensions'),
    'msix': DmapTag(read_bool,   'dmap.supportsindex'),
    'mslr': DmapTag(read_bool,   'dmap.loginrequired'),
    'mspi': DmapTag(read_bool,   'dmap.supportspersistentids'),
    'msqy': DmapTag(read_bool,   'dmap.supportsquery'),
    'msrv': DmapTag('container', 'dmap.serverinforesponse'),
    'mstc': DmapTag(read_uint,   'dmap.utctime'),
    'mstm': DmapTag(read_uint,   'dmap.timeoutinterval'),
    'msto': DmapTag(read_uint,   'dmap.utcoffset'),
    'mstt': DmapTag(read_uint,   'dmap.status'),
    'msup': DmapTag(read_bool,   'dmap.supportsupdate'),
    'mtco': DmapTag(read_uint,   'dmap.containercount'),

    # Tags with (yet) unknown purpose
    'aeFR': DmapTag(read_uint,   'unknown tag'),
    'aeSX': DmapTag(read_uint,   'unknown tag'),
    'asse': DmapTag(read_uint,   'unknown tag'),
    'atCV': DmapTag(read_uint,   'unknown tag'),
    'atSV': DmapTag(read_uint,   'unknown tag'),
    'caks': DmapTag(read_uint,   'unknown tag'),
    'caov': DmapTag(read_uint,   'unknown tag'),
    'casc': DmapTag(read_uint,   'unknown tag'),
    'cass': DmapTag(read_uint,   'unknown tag'),
    'ceQA': DmapTag(read_uint,   'unknown tag'),
    'cmbe': DmapTag(read_str,    'unknown tag'),
    'cmcc': DmapTag(read_str,    'unknown tag'),
    'cmce': DmapTag(read_str,    'unknown tag'),
    'cmcv': DmapTag(read_ignore, 'unknown tag'),
    'cmik': DmapTag(read_uint,   'unknown tag'),
    'cmsb': DmapTag(read_uint,   'unknown tag'),
    'cmsc': DmapTag(read_uint,   'unknown tag'),
    'cmsp': DmapTag(read_uint,   'unknown tag'),
    'cmsv': DmapTag(read_uint,   'unknown tag'),
    'cmte': DmapTag(read_str,    'unknown tag'),
    'mscu': DmapTag(read_uint,   'unknown tag'),
}


def lookup_tag(name):
    """Look up a tag based on its key. Returns a DmapTag."""
    return next((_TAGS[t] for t in _TAGS if t == name),
                DmapTag(_read_unknown, 'unknown tag'))



"""Smoke test for atvremote."""

import sys

from contextlib import contextmanager
from io import StringIO

from unittest.mock import patch

from aiohttp.test_utils import (AioHTTPTestCase, unittest_run_loop)

import pyatv
from pyatv import __main__ as atvremote
from tests import fake_udns, zeroconf_stub
from tests.utils import stub_sleep
from tests.airplay.fake_airplay_device import DEVICE_PIN, DEVICE_CREDENTIALS
from tests.mrp.fake_mrp_atv import (
    FakeAppleTV, AppleTVUseCases)
from tests.mrp.mrp_server_auth import CLIENT_CREDENTIALS


IP_1 = '10.0.0.1'
IP_2 = '127.0.0.1'
DMAP_ID = 'dmap_id'
MRP_ID = 'mrp_id'
AIRPLAY_ID = 'AA:BB:CC:DD:EE:FF'


@contextmanager
def capture_output(argv, inputs):
    new_out, new_err, new_in = StringIO(), StringIO(), StringIO(inputs)
    old_out, old_err, old_in = sys.stdout, sys.stderr, sys.stdin
    old_argv = sys.argv
    try:
        sys.stdout, sys.stderr, sys.stdin = new_out, new_err, new_in
        sys.argv = argv
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr, sys.stdin = old_out, old_err, old_in
        sys.argv = old_argv


class AtvremoteTest(AioHTTPTestCase):

    async def setUpAsync(self):
        await AioHTTPTestCase.setUpAsync(self)
        stub_sleep()
        self.setup_environment()
        await self.fake_udns.start()
        self.stdout = None
        self.stderr = None
        self.retcode = None
        self.inputs = []

    def setup_environment(self):
        airplay_port = self.server.port

        services = []
        services.append(zeroconf_stub.homesharing_service(
                DMAP_ID, b'Apple TV 1', IP_1, b'aaaa'))
        services.append(zeroconf_stub.mrp_service(
                'DDDD', b'Apple TV 2', IP_2, MRP_ID, port=self.fake_atv.port))
        services.append(zeroconf_stub.airplay_service(
                'Apple TV 2', IP_2, AIRPLAY_ID, port=airplay_port))
        zeroconf_stub.stub(pyatv, *services)

        self.fake_udns.add_service(fake_udns.mrp_service(
                'DDDD', 'Apple TV 2', MRP_ID, port=self.fake_atv.port))
        self.fake_udns.add_service(fake_udns.airplay_service(
                'Apple TV 2', AIRPLAY_ID, port=airplay_port))

        self.usecase.airplay_playback_playing()
        self.usecase.airplay_playback_idle()

    async def get_application(self, loop=None):
        self.fake_atv = FakeAppleTV(self, self.loop)
        self.usecase = AppleTVUseCases(self.fake_atv)
        self.fake_udns = fake_udns.FakeUdns(self.loop)
        return self.fake_atv.app

    def user_input(self, text):
        self.inputs.append(text)

    def has_output(self, *strings):
        for string in strings:
            self.assertIn(string, self.stdout)

    def has_error(self, *strings):
        for string in strings:
            self.assertIn(string, self.stderr)

    def exit(self, code):
        self.assertEqual(self.retcode, code)

    async def atvremote(self, *args):
        argv = ['atvremote'] + list(args)
        inputs = '\n'.join(self.inputs) + '\n'
        with capture_output(argv, inputs) as (out, err):
            udns_port = str(self.fake_udns.port)
            with patch.dict('os.environ', {'PYATV_UDNS_PORT': udns_port}):
                self.retcode = await atvremote.appstart(self.loop)
                self.stdout = out.getvalue()
                self.stderr = err.getvalue()

    @unittest_run_loop
    async def test_scan_devices(self):
        await self.atvremote("scan")
        self.has_output("Apple TV 1",
                        "Apple TV 2",
                        IP_1,
                        IP_2,
                        MRP_ID,
                        AIRPLAY_ID,
                        DMAP_ID)
        self.exit(0)

    @unittest_run_loop
    async def test_scan_hosts(self):
        await self.atvremote(
            "--scan-hosts",
            "127.0.0.1",
            "scan")
        self.has_output("Apple TV 2",
                        IP_2,
                        MRP_ID,
                        AIRPLAY_ID)
        self.exit(0)

    @unittest_run_loop
    async def test_pair_airplay(self):
        self.user_input(str(DEVICE_PIN))
        await self.atvremote(
            "--address", IP_2,
            "--protocol", "airplay",
            "--id", MRP_ID,
            "--airplay-credentials", DEVICE_CREDENTIALS,
            "pair")
        self.has_output("Enter PIN",
                        "seems to have succeeded",
                        DEVICE_CREDENTIALS)
        self.exit(0)

    @unittest_run_loop
    async def test_airplay_play_url(self):
        self.user_input(str(DEVICE_PIN))
        await self.atvremote(
            "--id", MRP_ID,
            "--airplay-credentials", DEVICE_CREDENTIALS,
            "play_url=http://fake")
        self.exit(0)

    @unittest_run_loop
    async def test_mrp_idle(self):
        await self.atvremote("--id", MRP_ID, "playing")
        self.has_output("Media type: Unknown", "Device state: Idle")
        self.exit(0)

    @unittest_run_loop
    async def test_mrp_auth(self):
        await self.atvremote(
            "--id", MRP_ID,
            "--mrp-credentials", CLIENT_CREDENTIALS,
            "playing")
        self.assertTrue(self.fake_atv.has_authenticated)
        self.has_output("Device state: Idle")
        self.exit(0)

    @unittest_run_loop
    async def test_mrp_auth_error(self):
        await self.atvremote(
            "--id", MRP_ID,
            "--mrp-credentials", "30:31:32:33",
            "playing")
        self.assertFalse(self.fake_atv.has_authenticated)
        self.has_error("AuthenticationError:")
        self.exit(1)

    @unittest_run_loop
    async def test_manual_connect(self):
        self.user_input(str(DEVICE_PIN))
        await self.atvremote(
            "--address", IP_2,
            "--protocol", "mrp",
            "--port", str(self.fake_atv.port),
            "--id", MRP_ID,
            "--manual",
            "playing")
        self.has_output("Media type: Unknown", "Device state: Idle")
        self.exit(0)



"""Unit tests for pyatv.convert."""

import unittest

from pyatv import (convert, exceptions)
from pyatv.const import (
    Protocol, MediaType, DeviceState, RepeatState, ShuffleState)

# These are extracted from iTunes, see for instance:
# http://www.blooming.no/wp-content/uploads/2013/03/ITLibMediaItem.h
# and also this:
# https://github.com/melloware/dacp-net/blob/master/Melloware.DACP/DACPResponse.cs
# Key: cmst.cmmk
MEDIA_KIND_UNKNOWN = 1
MEDIA_KIND_SONG = 2
MEDIA_KIND_MOVIE = 3
MEDIA_KIND_PODCAST = 4
MEDIA_KIND_AUDIOBOOK = 5
MEDIA_KIND_PDFBOOKLET = 6
MEDIA_KIND_MUSICVIDEO = 7
MEDIA_KIND_TVSHOW = 8
MEDIA_KIND_INTERACTIVEBOOKLET = 9
MEDIA_KIND_COACHEDAUDIO = 10
MEDIA_KIND_VIDEOPASS = 11
MEDIA_KIND_HOMEVIDEO = 12
MEDIA_KIND_FUTUREVIDEO = 13
MEDIA_KIND_RINGTONE = 14
MEDIA_KIND_DIGITALBOOKLET = 15
MEDIA_KIND_IOSAPPLICATION = 16
MEDIA_KIND_VOICEMEMO = 17
MEDIA_KIND_ITUNESU = 18
MEDIA_KIND_BOOK = 19
MEDIA_KIND_PDFBOOK = 20
MEDIA_KIND_ALERTTONE = 21
MEDIA_KIND_MUSICVIDEO2 = 32
MEDIA_KIND_PODCAST2 = 36
MEDIA_KIND_TVSHOW2 = 64

# Found on various places on the Internet as well as by testing
# Key: cmst.caps
PLAY_STATE_IDLE = 0
PLAY_STATE_LOADING = 1  # E.g. buffering
PLAY_STATE_STOPPED = 2
PLAY_STATE_PAUSED = 3
PLAY_STATE_PLAYING = 4
PLAY_STATE_FORWARD = 5
PLAY_STATE_BACKWARD = 6


class ConvertTest(unittest.TestCase):

    # MEDIA KIND TESTS

    def test_unknown_media_kind(self):
        self.assertEqual(MediaType.Unknown,
                         convert.media_kind(MEDIA_KIND_UNKNOWN))

    def test_video_media_kinds(self):
        self.assertEqual(MediaType.Video,
                         convert.media_kind(MEDIA_KIND_MOVIE))
        self.assertEqual(MediaType.Video,
                         convert.media_kind(MEDIA_KIND_MUSICVIDEO))
        self.assertEqual(MediaType.Video,
                         convert.media_kind(MEDIA_KIND_MUSICVIDEO2))
        self.assertEqual(MediaType.Video,
                         convert.media_kind(MEDIA_KIND_VIDEOPASS))
        self.assertEqual(MediaType.Video,
                         convert.media_kind(MEDIA_KIND_HOMEVIDEO))
        self.assertEqual(MediaType.Video,
                         convert.media_kind(MEDIA_KIND_FUTUREVIDEO))
        self.assertEqual(MediaType.Video,
                         convert.media_kind(MEDIA_KIND_ITUNESU))

    def test_music_media_kinds(self):
        self.assertEqual(MediaType.Music,
                         convert.media_kind(MEDIA_KIND_SONG))
        self.assertEqual(MediaType.Music,
                         convert.media_kind(MEDIA_KIND_PODCAST))
        self.assertEqual(MediaType.Music,
                         convert.media_kind(MEDIA_KIND_PODCAST2))
        self.assertEqual(MediaType.Music,
                         convert.media_kind(MEDIA_KIND_COACHEDAUDIO))
        self.assertEqual(MediaType.Music,
                         convert.media_kind(MEDIA_KIND_RINGTONE))
        self.assertEqual(MediaType.Music,
                         convert.media_kind(MEDIA_KIND_VOICEMEMO))
        self.assertEqual(MediaType.Music,
                         convert.media_kind(MEDIA_KIND_ALERTTONE))

    def test_tv_kinds(self):
        self.assertEqual(MediaType.TV,
                         convert.media_kind(MEDIA_KIND_TVSHOW))
        self.assertEqual(MediaType.TV,
                         convert.media_kind(MEDIA_KIND_TVSHOW2))

    def test_unknown_media_kind_throws(self):
        with self.assertRaises(exceptions.UnknownMediaKindError):
            convert.media_kind(99999)

    def test_media_type_to_string(self):
        self.assertEqual('Unknown',
                         convert.media_type_str(MediaType.Unknown))
        self.assertEqual('Video',
                         convert.media_type_str(MediaType.Video))
        self.assertEqual('Music',
                         convert.media_type_str(MediaType.Music))
        self.assertEqual('TV', convert.media_type_str(MediaType.TV))

    def test_unknown_media_type_to_str(self):
        self.assertEqual('Unsupported', convert.media_type_str(999))

    # PLAYSTATE TESTS

    def test_device_state_no_media(self):
        # This test should not really be here as "None" is in reality not a
        # valid value. But it is supported nonetheless because that makes
        # usage nicer. None means that the field is not included in a
        # server response, which matches the behavior of dmap.first.
        self.assertEqual(DeviceState.Idle,
                         convert.playstate(None))

    def test_regular_playstates(self):
        self.assertEqual(DeviceState.Idle,
                         convert.playstate(PLAY_STATE_IDLE))
        self.assertEqual(DeviceState.Loading,
                         convert.playstate(PLAY_STATE_LOADING))
        self.assertEqual(DeviceState.Stopped,
                         convert.playstate(PLAY_STATE_STOPPED))
        self.assertEqual(DeviceState.Paused,
                         convert.playstate(PLAY_STATE_PAUSED))
        self.assertEqual(DeviceState.Playing,
                         convert.playstate(PLAY_STATE_PLAYING))
        self.assertEqual(DeviceState.Seeking,
                         convert.playstate(PLAY_STATE_FORWARD))
        self.assertEqual(DeviceState.Seeking,
                         convert.playstate(PLAY_STATE_BACKWARD))

    def test_unknown_playstate_throws(self):
        with self.assertRaises(exceptions.UnknownPlayStateError):
            convert.playstate(99999)

    def test_playstate_str(self):
        self.assertEqual('Idle',
                         convert.playstate_str(DeviceState.Idle))
        self.assertEqual('Loading',
                         convert.playstate_str(DeviceState.Loading))
        self.assertEqual('Stopped',
                         convert.playstate_str(DeviceState.Stopped))
        self.assertEqual('Paused',
                         convert.playstate_str(DeviceState.Paused))
        self.assertEqual('Playing',
                         convert.playstate_str(DeviceState.Playing))
        self.assertEqual('Seeking',
                         convert.playstate_str(DeviceState.Seeking))

    def test_unsupported_playstate_str(self):
        self.assertEqual('Unsupported', convert.playstate_str(999))

    # TIME TESTS

    def test_no_time_returns_zero(self):
        self.assertEqual(0, convert.ms_to_s(None))

    def test_time_in_seconds(self):
        self.assertEqual(0, convert.ms_to_s(400))
        self.assertEqual(1, convert.ms_to_s(501))
        self.assertEqual(36, convert.ms_to_s(36000))

    def test_invalid_time(self):
        # Sometimes really large times are reported during buffering, this test
        # handles those special cases.
        self.assertEqual(0, convert.ms_to_s(2**32 - 1))

    # REPEAT TESTS

    def test_repeat_str(self):
        self.assertEqual('Off',
                         convert.repeat_str(RepeatState.Off))
        self.assertEqual('Track',
                         convert.repeat_str(RepeatState.Track))
        self.assertEqual('All',
                         convert.repeat_str(RepeatState.All))

    def test_unknown_repeat_to_str(self):
        self.assertEqual('Unsupported', convert.repeat_str(1234))

    # SHUFFLE TESTS

    def test_shuffle_str(self):
        self.assertEqual('Off',
                         convert.shuffle_str(ShuffleState.Off))
        self.assertEqual('Albums',
                         convert.shuffle_str(ShuffleState.Albums))
        self.assertEqual('Songs',
                         convert.shuffle_str(ShuffleState.Songs))

    def test_unknown_shuffle_to_str(self):
        self.assertEqual('Unsupported', convert.shuffle_str(1234))

    # PROTOCOL TESTS

    def test_protocol_str(self):
        self.assertEqual('MRP',
                         convert.protocol_str(Protocol.MRP))
        self.assertEqual('DMAP',
                         convert.protocol_str(Protocol.DMAP))
        self.assertEqual('AirPlay',
                         convert.protocol_str(Protocol.AirPlay))

    def test_unknown_protocol_str(self):
        self.assertEqual('Unknown', convert.protocol_str('invalid'))



"""Support functions used in library."""

import asyncio
import logging
import binascii

from pyatv import exceptions


async def error_handler(func, fallback, *args, **kwargs):
    """Call a function and re-map exceptions to match pyatv interface."""
    try:
        return await func(*args, **kwargs)
    except (OSError, asyncio.TimeoutError) as ex:
        raise exceptions.ConnectionFailedError(str(ex)) from ex
    except exceptions.BackOffError:
        raise
    except exceptions.NoCredentialsError:
        raise
    except Exception as ex:
        raise fallback(str(ex)) from ex


# Special log method to avoid hexlify conversion if debug is on
def log_binary(logger, message, **kwargs):
    """Log binary data if debug is enabled."""
    if logger.isEnabledFor(logging.DEBUG):
        output = ('{0}={1}'.format(k, binascii.hexlify(
            bytearray(v)).decode()) for k, v in sorted(kwargs.items()))
        logger.debug('%s (%s)', message, ', '.join(output))



"""Functional tests using the API with a fake Apple TV."""

from aiohttp.test_utils import unittest_run_loop

import pyatv
from pyatv.const import DeviceState, ShuffleState
from pyatv.conf import (AirPlayService, MrpService, AppleTV)

from tests import common_functional_tests
from tests.utils import until, faketime
from tests.mrp.fake_mrp_atv import (
    FakeAppleTV, AppleTVUseCases)
from tests.airplay.fake_airplay_device import DEVICE_CREDENTIALS


ARTWORK_BYTES = b'1234'
ARTWORK_MIMETYPE = 'image/png'
ARTWORK_ID = 'artwork_id1'


class MRPFunctionalTest(common_functional_tests.CommonFunctionalTests):

    async def setUpAsync(self):
        await super().setUpAsync()
        self.conf = AppleTV('127.0.0.1', 'Test device')
        self.conf.add_service(MrpService(
            'mrp_id', self.fake_atv.port))
        self.conf.add_service(AirPlayService(
            'airplay_id', self.server.port, DEVICE_CREDENTIALS))
        self.atv = await self.get_connected_device()

    async def tearDownAsync(self):
        await self.atv.close()
        await super().tearDownAsync()

    async def get_application(self, loop=None):
        self.fake_atv = FakeAppleTV(self, self.loop)
        self.usecase = AppleTVUseCases(self.fake_atv)
        return self.fake_atv.app

    async def get_connected_device(self):
        return await pyatv.connect(self.conf, loop=self.loop)

    @unittest_run_loop
    async def test_button_home(self):
        await self.atv.remote_control.home()
        await until(lambda: self.fake_atv.last_button_pressed == 'home')

    @unittest_run_loop
    async def test_button_volume_up(self):
        await self.atv.remote_control.volume_up()
        await until(lambda: self.fake_atv.last_button_pressed == 'volume_up')

    @unittest_run_loop
    async def test_button_volume_down(self):
        await self.atv.remote_control.volume_down()
        await until(lambda: self.fake_atv.last_button_pressed == 'volume_down')

    @unittest_run_loop
    async def test_button_suspend(self):
        await self.atv.remote_control.suspend()
        await until(lambda: self.fake_atv.last_button_pressed == 'suspend')

    @unittest_run_loop
    async def test_button_wakeup(self):
        await self.atv.remote_control.wakeup()
        await until(lambda: self.fake_atv.last_button_pressed == 'wakeup')

    @unittest_run_loop
    async def test_shuffle_state_albums(self):
        self.usecase.example_video(shuffle=ShuffleState.Albums)
        playing = await self.playing(shuffle=ShuffleState.Albums)
        self.assertEqual(playing.shuffle, ShuffleState.Albums)

    @unittest_run_loop
    async def test_set_shuffle_albums(self):
        self.usecase.example_video()

        await self.atv.remote_control.set_shuffle(ShuffleState.Albums)
        playing = await self.playing(shuffle=ShuffleState.Albums)
        self.assertEqual(playing.shuffle, ShuffleState.Albums)

    @unittest_run_loop
    async def test_metadata_artwork_id(self):
        self.usecase.example_video()
        self.usecase.change_artwork(
            ARTWORK_BYTES, ARTWORK_MIMETYPE, ARTWORK_ID)

        await self.playing(title='dummy')
        self.assertEqual(self.atv.metadata.artwork_id, ARTWORK_ID)

    @unittest_run_loop
    async def test_item_updates(self):
        self.usecase.video_playing(
            False, 'dummy', 100, 1, identifier='id', artist='some artist')

        with faketime('pyatv', 0):
            await self.playing(title='dummy')

            # Trigger update of single item by chaging title
            self.usecase.change_metadata(title='foobar', identifier='id')
            playing = await self.playing(title='foobar')

            # Make sure other metadata is untouched
            self.assertEqual(playing.title, 'foobar')
            self.assertEqual(playing.artist, 'some artist')
            self.assertEqual(playing.total_time, 100)
            self.assertEqual(playing.position, 1)

    @unittest_run_loop
    async def test_item_id_hash(self):
        initial_hash = (await self.atv.metadata.playing()).hash

        # Verify thar content identifier is used as hash
        self.usecase.example_video(identifier='some_id')
        playing = await self.playing(title='dummy')
        self.assertEqual(playing.hash, 'some_id')

        # Ensure that we fall back to initial hash if nothing is playing
        self.usecase.nothing_playing()
        nothing_playing = await self.playing(
            device_state=DeviceState.Idle)
        self.assertEqual(nothing_playing.hash, initial_hash)

    @unittest_run_loop
    async def test_metadata_playback_rate_device_state(self):
        self.usecase.example_video()

        playing = await self.playing(title='dummy')
        self.assertEqual(playing.device_state, DeviceState.Paused)

        self.usecase.change_metadata(title='dummy2', playback_rate=1.0)
        playing = await self.playing(title='dummy2')
        self.assertEqual(playing.device_state, DeviceState.Playing)



"""Unit tests for pyatv.dmap.parser."""

import unittest
import plistlib

from pyatv import exceptions
from pyatv.dmap import (tags, parser)

TEST_TAGS = {
    'uuu8': parser.DmapTag(tags.read_uint, 'uint8'),
    'uu16': parser.DmapTag(tags.read_uint, 'uint16'),
    'uu32': parser.DmapTag(tags.read_uint, 'uint32'),
    'uu64': parser.DmapTag(tags.read_uint, 'uint64'),
    'bola': parser.DmapTag(tags.read_bool, 'bool'),
    'bolb': parser.DmapTag(tags.read_bool, 'bool'),
    'stra': parser.DmapTag(tags.read_str, 'string'),
    'strb': parser.DmapTag(tags.read_str, 'string'),
    'cona': parser.DmapTag('container', 'container'),
    'conb': parser.DmapTag('container', 'container 2'),
    'igno': parser.DmapTag(tags.read_ignore, 'ignore'),
    'plst': parser.DmapTag(tags.read_bplist, 'bplist'),
}


def lookup_tag(name):
    return TEST_TAGS[name]


class ParserTest(unittest.TestCase):

    def test_parse_uint_of_various_lengths(self):
        in_data = tags.uint8_tag('uuu8', 12) + \
                  tags.uint16_tag('uu16', 37888) + \
                  tags.uint32_tag('uu32', 305419896) + \
                  tags.uint64_tag('uu64', 8982983289232)
        parsed = parser.parse(in_data, lookup_tag)
        self.assertEqual(4, len(parsed))
        self.assertEqual(12, parser.first(parsed, 'uuu8'))
        self.assertEqual(37888, parser.first(parsed, 'uu16'))
        self.assertEqual(305419896, parser.first(parsed, 'uu32'))
        self.assertEqual(8982983289232, parser.first(parsed, 'uu64'))

    def test_parse_bool(self):
        in_data = tags.bool_tag('bola', True) + \
                  tags.bool_tag('bolb', False)
        parsed = parser.parse(in_data, lookup_tag)
        self.assertEqual(2, len(parsed))
        self.assertTrue(parser.first(parsed, 'bola'))
        self.assertFalse(parser.first(parsed, 'bolb'))

    def test_parse_strings(self):
        in_data = tags.string_tag('stra', '') + \
                  tags.string_tag('strb', 'test string')
        parsed = parser.parse(in_data, lookup_tag)
        self.assertEqual(2, len(parsed))
        self.assertEqual('', parser.first(parsed, 'stra'))
        self.assertEqual('test string', parser.first(parsed, 'strb'))

    def test_parse_binary_plist(self):
        data = {"key": "value"}
        in_data = tags.raw_tag(
            'plst', plistlib.dumps(data, fmt=plistlib.FMT_BINARY))
        parsed = parser.parse(in_data, lookup_tag)
        self.assertEqual(1, len(parsed))
        self.assertEqual(data, parser.first(parsed, 'plst'))

    def test_parse_value_in_container(self):
        in_data = tags.container_tag('cona',
                                     tags.uint8_tag('uuu8', 36) +
                                     tags.uint16_tag('uu16', 13000))
        parsed = parser.parse(in_data, lookup_tag)
        self.assertEqual(1, len(parsed))
        inner = parser.first(parsed, 'cona')
        self.assertEqual(2, len(inner))
        self.assertEqual(36, parser.first(inner, 'uuu8'))
        self.assertEqual(13000, parser.first(inner, 'uu16'))

    def test_extract_simplified_container(self):
        elem = tags.uint8_tag('uuu8', 12)
        inner = tags.container_tag('conb', elem)
        in_data = tags.container_tag('cona', inner)
        parsed = parser.parse(in_data, lookup_tag)
        self.assertEqual(12, parser.first(parsed, 'cona', 'conb', 'uuu8'))

    def test_ignore_value(self):
        elem = tags.uint8_tag('igno', 44)
        parsed = parser.parse(elem, lookup_tag)
        self.assertEqual(parser.first(parsed, 'igno'), None)

    def test_simple_pprint(self):
        elem = tags.uint8_tag('uuu8', 12)
        inner = tags.container_tag('conb', elem)
        in_data = tags.container_tag('cona', inner)
        parsed = parser.parse(in_data, lookup_tag)
        self.assertEqual(parser.pprint(parsed, lookup_tag),
                         'cona: [container, container]\n' +
                         '  conb: [container, container 2]\n' +
                         '    uuu8: 12 [uint, uint8]\n')

    def test_print_invalid_input_raises_exception(self):
        with self.assertRaises(exceptions.InvalidDmapDataError):
            parser.pprint('bad data', lookup_tag)



"""Functional pairing tests using the API with a fake AirPlay Apple TV."""

import binascii
from asynctest.mock import patch

from aiohttp.test_utils import (AioHTTPTestCase, unittest_run_loop)

from pyatv import pair, exceptions
from pyatv.const import Protocol
from pyatv.conf import (AirPlayService, AppleTV)
from tests.airplay.fake_airplay_device import (
    FakeAirPlayDevice, AirPlayUseCases, DEVICE_CREDENTIALS, DEVICE_PIN,
    DEVICE_IDENTIFIER, DEVICE_AUTH_KEY)


def predetermined_key(num):
    """Return random data corresponding to hardcoded AirPlay keys."""
    if num == 8:
        return binascii.unhexlify(DEVICE_IDENTIFIER)
    return binascii.unhexlify(DEVICE_AUTH_KEY)


class PairFunctionalTest(AioHTTPTestCase):

    def setUp(self):
        AioHTTPTestCase.setUp(self)
        self.pairing = None

        self.service = AirPlayService(
            'airplay_id', port=self.server.port)
        self.conf = AppleTV('127.0.0.1', 'Apple TV')
        self.conf.add_service(self.service)

    async def tearDownAsync(self):
        await self.pairing.close()
        await super().tearDownAsync()

    async def get_application(self, loop=None):
        self.fake_atv = FakeAirPlayDevice(self)
        self.usecase = AirPlayUseCases(self.fake_atv)
        return self.fake_atv.app

    async def do_pairing(self, pin=DEVICE_PIN):
        self.usecase.airplay_require_authentication()

        self.pairing = await pair(
            self.conf, Protocol.AirPlay, self.loop)

        self.assertTrue(self.pairing.device_provides_pin)

        await self.pairing.begin()
        if pin:
            self.pairing.pin(pin)

        self.assertFalse(self.pairing.has_paired)

        await self.pairing.finish()
        self.assertTrue(self.pairing.has_paired)
        self.assertEqual(self.service.credentials, DEVICE_CREDENTIALS)

    @unittest_run_loop
    async def test_pairing_exception_invalid_pin(self):
        with self.assertRaises(exceptions.PairingError):
            await self.do_pairing(9999)

    @unittest_run_loop
    async def test_pairing_exception_no_pin(self):
        with self.assertRaises(exceptions.PairingError):
            await self.do_pairing(None)

    @unittest_run_loop
    @patch('os.urandom')
    async def test_pairing_with_device_new_credentials(self, rand_func):
        rand_func.side_effect = predetermined_key
        await self.do_pairing()

    @unittest_run_loop
    async def test_pairing_with_device_existing_credentials(self):
        self.conf.get_service(
            Protocol.AirPlay).credentials = DEVICE_CREDENTIALS
        await self.do_pairing()



"""Simple example showing of pairing."""

import sys
import asyncio

from pyatv import scan, pair
from pyatv.const import Protocol


LOOP = asyncio.get_event_loop()


# Method that is dispatched by the asyncio event loop
async def pair_with_device(loop):
    """Make it possible to pair with device."""
    atvs = await scan(loop, timeout=5, protocol=Protocol.MRP)

    if not atvs:
        print('No device found', file=sys.stderr)
        return

    pairing = await pair(atvs[0], Protocol.MRP, loop)
    await pairing.begin()

    pin = int(input("Enter PIN: "))
    pairing.pin(pin)
    await pairing.finish()

    # Give some feedback about the process
    if pairing.has_paired:
        print('Paired with device!')
        print('Credentials:', pairing.service.credentials)
    else:
        print('Did not pair with device!')

    await pairing.close()

if __name__ == '__main__':
    # Setup event loop and connect
    LOOP.run_until_complete(pair_with_device(LOOP))



"""Unit tests for pyatv.conf."""

import unittest

from pyatv import (conf, exceptions)
from pyatv.const import Protocol

ADDRESS_1 = '127.0.0.1'
ADDRESS_2 = '192.168.0.1'
NAME = 'Alice'
PORT_1 = 1234
PORT_2 = 5678
IDENTIFIER_1 = 'id1'
IDENTIFIER_2 = 'id2'
IDENTIFIER_3 = 'id3'
CREDENTIALS_1 = 'cred1'


class ConfTest(unittest.TestCase):

    def setUp(self):
        self.config = conf.AppleTV(ADDRESS_1, NAME)
        self.dmap_service = conf.DmapService(
            IDENTIFIER_1, None, port=PORT_1)
        self.mrp_service = conf.MrpService(IDENTIFIER_2, PORT_2)
        self.airplay_service = conf.AirPlayService(
            IDENTIFIER_3, PORT_1)

    def test_address_and_name(self):
        self.assertEqual(self.config.address, ADDRESS_1)
        self.assertEqual(self.config.name, NAME)

    def test_equality(self):
        self.assertEqual(self.config, self.config)

        atv2 = conf.AppleTV(ADDRESS_1, NAME)
        atv2.add_service(conf.AirPlayService(IDENTIFIER_1, PORT_1))
        self.assertNotEqual(self.config, atv2)

    def test_add_services_and_get(self):
        self.config.add_service(self.dmap_service)
        self.config.add_service(self.mrp_service)
        self.config.add_service(self.airplay_service)

        services = self.config.services
        self.assertEqual(len(services), 3)

        self.assertIn(self.dmap_service, services)
        self.assertIn(self.mrp_service, services)
        self.assertIn(self.airplay_service, services)

        self.assertEqual(
            self.config.get_service(Protocol.DMAP), self.dmap_service)
        self.assertEqual(
            self.config.get_service(Protocol.MRP), self.mrp_service)
        self.assertEqual(
            self.config.get_service(Protocol.AirPlay), self.airplay_service)

    def test_identifier_order(self):
        self.assertIsNone(self.config.identifier)

        self.config.add_service(self.dmap_service)
        self.assertEqual(self.config.identifier, IDENTIFIER_1)

        self.config.add_service(self.mrp_service)
        self.assertEqual(self.config.identifier, IDENTIFIER_2)

        self.config.add_service(self.airplay_service)
        self.assertEqual(self.config.identifier, IDENTIFIER_2)

    def test_add_airplay_service(self):
        self.config.add_service(self.airplay_service)

        airplay = self.config.get_service(Protocol.AirPlay)
        self.assertEqual(airplay.protocol, Protocol.AirPlay)
        self.assertEqual(airplay.port, PORT_1)

    def test_main_service_no_service(self):
        with self.assertRaises(exceptions.NoServiceError):
            self.config.main_service()

    def test_main_service_airplay_no_service(self):
        self.config.add_service(self.airplay_service)
        with self.assertRaises(exceptions.NoServiceError):
            self.config.main_service()

    def test_main_service_get_service(self):
        self.config.add_service(self.dmap_service)
        self.assertEqual(self.config.main_service(), self.dmap_service)

        self.config.add_service(self.mrp_service)
        self.assertEqual(self.config.main_service(), self.mrp_service)

    def test_main_service_override_protocol(self):
        self.config.add_service(self.dmap_service)
        self.config.add_service(self.mrp_service)
        self.assertEqual(
            self.config.main_service(protocol=self.dmap_service.protocol),
            self.dmap_service)

    def test_set_credentials_for_missing_service(self):
        self.assertFalse(self.config.set_credentials(Protocol.DMAP, 'dummy'))

    def test_set_credentials(self):
        self.config.add_service(self.dmap_service)
        self.assertIsNone(self.config.get_service(Protocol.DMAP).credentials)

        self.config.set_credentials(Protocol.DMAP, 'dummy')
        self.assertEqual(
            self.config.get_service(Protocol.DMAP).credentials, 'dummy')

    # This test is a bit strange and couples to protocol specific services,
    # but it's mainly to exercise string as that is important. Might refactor
    # this in the future.
    def test_to_str(self):
        self.config.add_service(conf.DmapService(IDENTIFIER_1, 'LOGIN_ID'))
        self.config.add_service(conf.MrpService(IDENTIFIER_2, PORT_2))

        # Check for some keywords to not lock up format too much
        output = str(self.config)
        self.assertIn(ADDRESS_1, output)
        self.assertIn(NAME, output)
        self.assertIn('LOGIN_ID', output)
        self.assertIn(str(PORT_2), output)
        self.assertIn('3689', output)



"""Util functions for extracting and constructing DMAP data."""

import plistlib


def read_str(data, start, length):
    """Extract a string from a position in a sequence."""
    return data[start:start+length].decode('utf-8')


def read_uint(data, start, length):
    """Extract a uint from a position in a sequence."""
    return int.from_bytes(data[start:start+length], byteorder='big')


def read_bool(data, start, length):
    """Extract a boolean from a position in a sequence."""
    return read_uint(data, start, length) == 1


def read_bplist(data, start, length):
    """Extract a binary plist from a position in a sequence."""
    # TODO: pylint doesn't find FMT_BINARY, why?
    # pylint: disable=no-member
    return plistlib.loads(data[start:start+length],
                          fmt=plistlib.FMT_BINARY)


# pylint: disable=unused-argument
def read_ignore(data, start, length):
    """Use this to ignore data for all input."""


def uint8_tag(name, value):
    """Create a DMAP tag with uint8 data."""
    return name.encode('utf-8') + \
        b'\x00\x00\x00\x01' + \
        value.to_bytes(1, byteorder='big')


def uint16_tag(name, value):
    """Create a DMAP tag with uint16 data."""
    return name.encode('utf-8') + \
        b'\x00\x00\x00\x02' + \
        value.to_bytes(2, byteorder='big')


def uint32_tag(name, value):
    """Create a DMAP tag with uint32 data."""
    return name.encode('utf-8') + \
        b'\x00\x00\x00\x04' + \
        value.to_bytes(4, byteorder='big')


def uint64_tag(name, value):
    """Create a DMAP tag with uint64 data."""
    return name.encode('utf-8') + \
        b'\x00\x00\x00\x08' + \
        value.to_bytes(8, byteorder='big')


def bool_tag(name, value):
    """Create a DMAP tag with boolean data."""
    return name.encode('utf-8') + \
        b'\x00\x00\x00\x01' + \
        (b'\x01' if value else b'\x00')


def raw_tag(name, value):
    """Create a DMAP tag with raw data."""
    return name.encode('utf-8') + \
        len(value).to_bytes(4, byteorder='big') + \
        value


def string_tag(name, value):
    """Create a DMAP tag with string data."""
    return name.encode('utf-8') + \
        len(value).to_bytes(4, byteorder='big') + \
        value.encode('utf-8')


def container_tag(name, data):
    """Create a DMAP tag with string data."""
    return raw_tag(name, data)  # Same as raw



"""API exposed by the library."""

import re
import inspect
import hashlib

from collections import namedtuple
from abc import (ABCMeta, abstractmethod, abstractproperty)

from pyatv import (convert, exceptions, net)


ArtworkInfo = namedtuple('ArtworkInfo', 'bytes mimetype')


def _get_first_sentence_in_pydoc(obj):
    doc = obj.__doc__
    index = doc.find('.')
    if index == -1:
        # Here we have no leading . so return everything
        return doc

    # Try to find the first complete sentence and respect
    # abbreviations correctly
    match = re.findall(r'(.*\.[^A-Z]*)\.(?: [A-Z].*|)', doc)
    if len(match) == 1:
        return match[0]
    return doc[0:index]


def retrieve_commands(obj):
    """Retrieve all commands and help texts from an API object."""
    commands = {}  # Name and help
    for func in obj.__dict__:
        if not inspect.isfunction(obj.__dict__[func]) and \
           not isinstance(obj.__dict__[func], property):
            continue
        if func.startswith('_'):
            continue
        commands[func] = _get_first_sentence_in_pydoc(
            obj.__dict__[func])
    return commands


class PairingHandler:
    """Base class for API used to pair with an Apple TV."""

    __metaclass__ = ABCMeta

    def __init__(self, session, service):
        """Initialize a new instance of PairingHandler."""
        self.session = session
        self._service = service

    @property
    def service(self):
        """Return service used for pairing."""
        return self._service

    async def close(self):
        """Call to free allocated resources after pairing."""
        if net.is_custom_session(self.session):
            await self.session.close()

    @abstractmethod
    def pin(self, pin):
        """Pin code used for pairing."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def device_provides_pin(self):
        """Return True if remote device presents PIN code, else False."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def has_paired(self):
        """If a successful pairing has been performed.

        The value will be reset when stop() is called.
        """
        raise exceptions.NotSupportedError()

    @abstractmethod
    async def begin(self):
        """Start pairing process."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    async def finish(self):
        """Stop pairing process."""
        raise exceptions.NotSupportedError()


class RemoteControl:  # pylint: disable=too-many-public-methods
    """Base class for API used to control an Apple TV."""

    __metaclass__ = ABCMeta

    # pylint: disable=invalid-name
    @abstractmethod
    def up(self):
        """Press key up."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def down(self):
        """Press key down."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def left(self):
        """Press key left."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def right(self):
        """Press key right."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def play(self):
        """Press key play."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def pause(self):
        """Press key play."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def stop(self):
        """Press key stop."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def next(self):
        """Press key next."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def previous(self):
        """Press key previous."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def select(self):
        """Press key select."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def menu(self):
        """Press key menu."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def volume_up(self):
        """Press key volume up."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def volume_down(self):
        """Press key volume down."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def home(self):
        """Press key home."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def home_hold(self):
        """Hold key home."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def top_menu(self):
        """Go to main menu (long press menu)."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def suspend(self):
        """Suspend the device."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def wakeup(self):
        """Wake up the device."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def set_position(self, pos):
        """Seek in the current playing media."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def set_shuffle(self, shuffle_state):
        """Change shuffle mode to on or off."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def set_repeat(self, repeat_state):
        """Change repeat state."""
        raise exceptions.NotSupportedError()


class Playing:
    """Base class for retrieving what is currently playing."""

    __metaclass__ = ABCMeta

    def __str__(self):
        """Convert this playing object to a readable string."""
        output = []
        output.append('  Media type: {0}'.format(
            convert.media_type_str(self.media_type)))
        output.append('Device state: {0}'.format(
            convert.playstate_str(self.device_state)))

        if self.title is not None:
            output.append('       Title: {0}'.format(self.title))

        if self.artist is not None:
            output.append('      Artist: {0}'.format(self.artist))

        if self.album is not None:
            output.append('       Album: {0}'.format(self.album))

        if self.genre is not None:
            output.append('       Genre: {0}'.format(self.genre))

        position = self.position
        total_time = self.total_time
        if position is not None and total_time is not None and total_time != 0:
            output.append('    Position: {0}/{1}s ({2:.1%})'.format(
                position, total_time, float(position)/float(total_time)))
        elif position is not None and position != 0:
            output.append('    Position: {0}s'.format(position))
        elif total_time is not None and position != 0:
            output.append('  Total time: {0}s'.format(total_time))

        if self.repeat is not None:
            output.append('      Repeat: {0}'.format(
                convert.repeat_str(self.repeat)))

        if self.shuffle is not None:
            output.append('     Shuffle: {0}'.format(
                convert.shuffle_str(self.shuffle)))

        return '\n'.join(output)

    @property
    def hash(self):
        """Create a unique hash for what is currently playing.

        The hash is based on title, artist, album and total time. It should
        always be the same for the same content, but it is not guaranteed.
        """
        base = '{0}{1}{2}{3}'.format(
            self.title, self.artist, self.album, self.total_time)
        return hashlib.sha256(base.encode('utf-8')).hexdigest()

    @abstractproperty
    def media_type(self):
        """Type of media is currently playing, e.g. video, music."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def device_state(self):
        """Device state, e.g. playing or paused."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def title(self):
        """Title of the current media, e.g. movie or song name."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def artist(self):
        """Artist of the currently playing song."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def album(self):
        """Album of the currently playing song."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def genre(self):
        """Genre of the currently playing song."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def total_time(self):
        """Total play time in seconds."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def position(self):
        """Position in the playing media (seconds)."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def shuffle(self):
        """If shuffle is enabled or not."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def repeat(self):
        """Repeat mode."""
        raise exceptions.NotSupportedError()


class Metadata:
    """Base class for retrieving metadata from an Apple TV."""

    __metaclass__ = ABCMeta

    def __init__(self, identifier):
        """Initialize a new instance of Metadata."""
        self._identifier = identifier

    @property
    def device_id(self):
        """Return a unique identifier for current device."""
        return self._identifier

    @abstractmethod
    def artwork(self):
        """Return artwork for what is currently playing (or None)."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def artwork_id(self):
        """Return a unique identifier for current artwork."""
        raise exceptions.NotSupportedError()

    @abstractmethod
    def playing(self):
        """Return what is currently playing."""
        raise exceptions.NotSupportedError()


class PushUpdater:
    """Base class for push/async updates from an Apple TV."""

    __metaclass__ = ABCMeta

    def __init__(self):
        """Initialize a new PushUpdater."""
        self.__listener = None

    @abstractproperty
    def active(self):
        """Return if push updater has been started."""
        raise exceptions.NotSupportedError()

    @property
    def listener(self):
        """Object (PushUpdaterListener) that receives updates."""
        return self.__listener

    @listener.setter  # type: ignore
    def listener(self, listener):
        """Object that receives updates.

        This should be an object implementing two methods:
        - playstatus_update(updater, playstatus)
        - playstatus_error(updater, exception)

        The first method is called when a new update happens and the second one
        is called if an error occurs.
        """
        self.__listener = listener

    @abstractmethod
    def start(self, initial_delay=0):
        """Begin to listen to updates.

        If an error occurs, start must be called again.
        """
        raise exceptions.NotSupportedError()

    @abstractmethod
    def stop(self):
        """No longer forward updates to listener."""
        raise exceptions.NotSupportedError()


class Stream:  # pylint: disable=too-few-public-methods
    """Base class for stream functionality."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def play_url(self, url, **kwargs):
        """Play media from an URL on the device."""
        raise exceptions.NotSupportedError()


class DeviceListener:
    """Listener interface for generic device updates."""

    @abstractmethod
    def connection_lost(self, exception):
        """Device was unexpectedly disconnected."""
        raise NotImplementedError()

    @abstractmethod
    def connection_closed(self):
        """Device connection was (intentionally) closed."""
        raise NotImplementedError()


class AppleTV:
    """Base class representing an Apple TV."""

    __metaclass__ = ABCMeta

    def __init__(self):
        """Initialize a new AppleTV."""
        self.__listener = None

    @property
    def listener(self):
        """Object receiving generic device updates.

        Must be an object conforming to DeviceListener.
        """
        return self.__listener

    @listener.setter
    def listener(self, target):
        """Change object receiving generic device updates."""
        self.__listener = target

    @abstractmethod
    async def connect(self):
        """Initiate connection to device.

        No need to call it yourself, it's done automatically.
        """
        raise exceptions.NotSupportedError()

    @abstractmethod
    async def close(self):
        """Close connection and release allocated resources."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def service(self):
        """Return service used to connect to the Apple TV."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def remote_control(self):
        """Return API for controlling the Apple TV."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def metadata(self):
        """Return API for retrieving metadata from the Apple TV."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def push_updater(self):
        """Return API for handling push update from the Apple TV."""
        raise exceptions.NotSupportedError()

    @abstractproperty
    def stream(self):
        """Return API for streaming media."""
        raise exceptions.NotSupportedError()
